#!/usr/bin/env python3
"""
Python Language Queries

Comprehensive Tree-sitter queries for Python language constructs.
Covers functions, classes, variables, imports, decorators, async/await,
type hints, and modern Python features.
Equivalent to JavaScript query coverage for consistent language support.
"""

# Function definitions
FUNCTIONS = """
(function_definition
    name: (identifier) @function.name
    parameters: (parameters) @function.params
    body: (block) @function.body) @function.definition

(function_definition
    name: (identifier) @function.name
    parameters: (parameters) @function.params
    body: (block) @function.body) @function.async
"""

# Class definitions
CLASSES = """
(class_definition
    name: (identifier) @class.name
    superclasses: (argument_list)? @class.superclasses
    body: (block) @class.body) @class.definition
"""

# Import statements
IMPORTS = """
(import_statement
    name: (dotted_name) @import.name) @import.statement

(import_from_statement
    module_name: (dotted_name)? @import.module
    name: (dotted_name) @import.name) @import.from

(import_from_statement
    module_name: (dotted_name)? @import.module
    name: (aliased_import) @import.aliased_item) @import.from_aliased

(aliased_import
    name: (dotted_name) @import.name
    alias: (identifier) @import.alias) @import.aliased
"""

# Variable assignments
VARIABLES = """
(assignment
    left: (identifier) @variable.name
    right: (_) @variable.value) @variable.assignment

(assignment
    left: (pattern_list) @variable.pattern
    right: (_) @variable.value) @variable.multiple

(augmented_assignment
    left: (identifier) @variable.name
    right: (_) @variable.value) @variable.augmented
"""

# Decorators
DECORATORS = """
(decorator
    (identifier) @decorator.name) @decorator.simple

(decorator
    (call
        function: (identifier) @decorator.name
        arguments: (argument_list) @decorator.args)) @decorator.call

(decorator
    (attribute
        object: (identifier) @decorator.object
        attribute: (identifier) @decorator.name)) @decorator.attribute
"""

# Method definitions
METHODS = """
(function_definition
    name: (identifier) @method.name
    parameters: (parameters
        (identifier) @method.self
        . (_)*) @method.params
    body: (block) @method.body) @method.definition
"""

# Exception handling
EXCEPTIONS = """
(try_statement
    body: (block) @try.body
    (except_clause
        type: (_)? @except.type
        name: (identifier)? @except.name
        body: (block) @except.body)*
    (else_clause
        body: (block) @else.body)?
    (finally_clause
        body: (block) @finally.body)?) @try.statement

(raise_statement
    (call
        function: (identifier) @exception.name
        arguments: (argument_list)? @exception.args)) @raise.statement
"""

# Comprehensions
COMPREHENSIONS = """
(list_comprehension
    body: (_) @comprehension.body
    (for_in_clause
        left: (_) @comprehension.var
        right: (_) @comprehension.iter)) @list.comprehension

(dictionary_comprehension
    body: (pair
        key: (_) @comprehension.key
        value: (_) @comprehension.value)
    (for_in_clause
        left: (_) @comprehension.var
        right: (_) @comprehension.iter)) @dict.comprehension

(set_comprehension
    body: (_) @comprehension.body
    (for_in_clause
        left: (_) @comprehension.var
        right: (_) @comprehension.iter)) @set.comprehension
"""

# Comments and docstrings
COMMENTS = """
(comment) @comment

(expression_statement
    (string) @docstring)
"""

# Type hints and annotations
TYPE_HINTS = """
(function_definition
    parameters: (parameters
        (typed_parameter
            type: (_) @type.param)) @type.function_param)

(function_definition
    return_type: (_) @type.return) @type.function_return

(assignment
    type: (_) @type.variable) @type.variable_annotation
"""

# Async/await patterns
ASYNC_PATTERNS = """
(function_definition) @async.function

(await
    (call) @async.await_call) @async.await

(async_for_statement) @async.for

(async_with_statement) @async.with
"""

# F-strings and string formatting
STRING_FORMATTING = """
(formatted_string
    (interpolation) @string.interpolation) @string.fstring

(call
    function: (attribute
        object: (_)
        attribute: (identifier) @string.format_method))
"""

# Context managers
CONTEXT_MANAGERS = """
(with_statement
    (with_clause
        (with_item
            value: (_) @context.manager)) @context.clause) @context.with

(async_with_statement
    (with_clause
        (with_item
            value: (_) @context.manager)) @context.clause) @context.async_with
"""

# Lambda expressions
LAMBDAS = """
(lambda
    parameters: (lambda_parameters)? @lambda.params
    body: (_) @lambda.body) @lambda.expression
"""

# Modern Python patterns
MODERN_PATTERNS = """
(match_statement
    subject: (_) @match.subject
    body: (case_clause)+ @match.cases) @pattern.match

(case_clause
    pattern: (_) @case.pattern
    guard: (_)? @case.guard
    consequence: (block) @case.body) @pattern.case

(walrus_operator
    left: (_) @walrus.target
    right: (_) @walrus.value) @assignment.walrus
"""

# Python-specific comprehensive query library
PYTHON_QUERIES: dict[str, str] = {
    # --- Basic Structure ---
    "function": """
    (function_definition) @function
    """,
    "function_definition": """
    (function_definition
        name: (identifier) @function_name
        parameters: (parameters) @parameters
        body: (block) @body) @function_definition
    """,
    "async_function": """
    (function_definition
        "async" @async_keyword
        name: (identifier) @function_name
        parameters: (parameters) @parameters
        body: (block) @body) @async_function
    """,
    "method": """
    (function_definition
        name: (identifier) @method_name
        parameters: (parameters
            (identifier) @self_param
            . (_)*) @method_params
        body: (block) @method_body) @method_definition
    """,
    "lambda": """
    (lambda
        parameters: (lambda_parameters)? @lambda_params
        body: (_) @lambda_body) @lambda_expression
    """,
    # --- Classes ---
    "class": """
    (class_definition) @class
    """,
    "class_definition": """
    (class_definition
        name: (identifier) @class_name
        superclasses: (argument_list)? @superclasses
        body: (block) @body) @class_definition
    """,
    "class_method": """
    (class_definition
        body: (block
            (function_definition
                name: (identifier) @method_name
                parameters: (parameters) @parameters
                body: (block) @method_body))) @class_method
    """,
    "constructor": """
    (function_definition
        name: (identifier) @constructor_name
        (#match? @constructor_name "__init__")
        parameters: (parameters) @parameters
        body: (block) @body) @constructor
    """,
    "property": """
    (decorated_definition
        (decorator
            (identifier) @decorator_name
            (#match? @decorator_name "property"))
        (function_definition
            name: (identifier) @property_name)) @property_definition
    """,
    "staticmethod": """
    (decorated_definition
        (decorator
            (identifier) @decorator_name
            (#match? @decorator_name "staticmethod"))
        (function_definition
            name: (identifier) @method_name)) @static_method
    """,
    "classmethod": """
    (decorated_definition
        (decorator
            (identifier) @decorator_name
            (#match? @decorator_name "classmethod"))
        (function_definition
            name: (identifier) @method_name)) @class_method_decorator
    """,
    # --- Variables and Assignments ---
    "variable": """
    (assignment) @variable
    """,
    "assignment": """
    (assignment
        left: (identifier) @variable_name
        right: (_) @value) @assignment
    """,
    "multiple_assignment": """
    (assignment
        left: (pattern_list) @variables
        right: (_) @value) @multiple_assignment
    """,
    "augmented_assignment": """
    (augmented_assignment
        left: (identifier) @variable_name
        right: (_) @value) @augmented_assignment
    """,
    "global_statement": """
    (global_statement
        (identifier) @global_var) @global_declaration
    """,
    "nonlocal_statement": """
    (nonlocal_statement
        (identifier) @nonlocal_var) @nonlocal_declaration
    """,
    # --- Imports ---
    "import": """
    (import_statement) @import
    """,
    "import_statement": """
    (import_statement
        name: (dotted_name) @import_name) @import_statement
    """,
    "import_from": """
    (import_from_statement
        module_name: (dotted_name)? @module_name
        name: (dotted_name) @import_name) @import_from
    """,
    "import_from_list": """
    (import_from_statement
        module_name: (dotted_name)? @module_name
        name: (aliased_import) @import_item) @import_from_list
    """,
    "aliased_import": """
    (aliased_import
        name: (dotted_name) @import_name
        alias: (identifier) @alias) @aliased_import
    """,
    "import_star": """
    (import_from_statement
        module_name: (dotted_name) @module_name
        name: (wildcard_import) @star_import) @import_star
    """,
    # --- Decorators ---
    "decorator": """
    (decorator) @decorator
    """,
    "decorator_simple": """
    (decorator
        (identifier) @decorator_name) @decorator_simple
    """,
    "decorator_call": """
    (decorator
        (call
            function: (identifier) @decorator_name
            arguments: (argument_list) @decorator_args)) @decorator_call
    """,
    "decorator_attribute": """
    (decorator
        (attribute
            object: (identifier) @decorator_object
            attribute: (identifier) @decorator_name)) @decorator_attribute
    """,
    "decorated_function": """
    (decorated_definition
        (decorator)+ @decorators
        (function_definition
            name: (identifier) @function_name)) @decorated_function
    """,
    "decorated_class": """
    (decorated_definition
        (decorator)+ @decorators
        (class_definition
            name: (identifier) @class_name)) @decorated_class
    """,
    # --- Control Flow ---
    "if_statement": """
    (if_statement
        condition: (_) @condition
        consequence: (block) @then_branch
        alternative: (_)? @else_branch) @if_statement
    """,
    "for_statement": """
    (for_statement
        left: (_) @loop_var
        right: (_) @iterable
        body: (block) @body) @for_statement
    """,
    "while_statement": """
    (while_statement
        condition: (_) @condition
        body: (block) @body) @while_statement
    """,
    "with_statement": """
    (with_statement
        (with_clause
            (with_item
                value: (_) @context_manager)) @with_clause
        body: (block) @body) @with_statement
    """,
    "async_with": """
    (async_with_statement
        (with_clause
            (with_item
                value: (_) @context_manager)) @with_clause
        body: (block) @body) @async_with_statement
    """,
    "async_for": """
    (async_for_statement
        left: (_) @loop_var
        right: (_) @async_iterable
        body: (block) @body) @async_for_statement
    """,
    # --- Exception Handling ---
    "try_statement": """
    (try_statement
        body: (block) @try_body
        (except_clause)* @except_clauses
        (else_clause)? @else_clause
        (finally_clause)? @finally_clause) @try_statement
    """,
    "except_clause": """
    (except_clause
        type: (_)? @exception_type
        name: (identifier)? @exception_name
        body: (block) @except_body) @except_clause
    """,
    "raise_statement": """
    (raise_statement
        (call
            function: (identifier) @exception_name
            arguments: (argument_list)? @exception_args)) @raise_statement
    """,
    "assert_statement": """
    (assert_statement
        (_) @assertion
        (_)? @message) @assert_statement
    """,
    # --- Comprehensions ---
    "list_comprehension": """
    (list_comprehension
        body: (_) @comprehension_body
        (for_in_clause
            left: (_) @comprehension_var
            right: (_) @comprehension_iter)) @list_comprehension
    """,
    "dict_comprehension": """
    (dictionary_comprehension
        body: (pair
            key: (_) @comprehension_key
            value: (_) @comprehension_value)
        (for_in_clause
            left: (_) @comprehension_var
            right: (_) @comprehension_iter)) @dict_comprehension
    """,
    "set_comprehension": """
    (set_comprehension
        body: (_) @comprehension_body
        (for_in_clause
            left: (_) @comprehension_var
            right: (_) @comprehension_iter)) @set_comprehension
    """,
    "generator_expression": """
    (generator_expression
        body: (_) @generator_body
        (for_in_clause
            left: (_) @generator_var
            right: (_) @generator_iter)) @generator_expression
    """,
    # --- Type Hints and Annotations ---
    "type_hint": """
    (typed_parameter
        type: (_) @parameter_type) @typed_parameter
    """,
    "return_type": """
    (function_definition
        return_type: (_) @return_type) @function_with_return_type
    """,
    "variable_annotation": """
    (assignment
        type: (_) @variable_type) @annotated_assignment
    """,
    "type_alias": """
    (type_alias_statement
        name: (identifier) @alias_name
        value: (_) @alias_type) @type_alias
    """,
    # --- Modern Python Features ---
    "match_statement": """
    (match_statement
        subject: (_) @match_subject
        body: (case_clause)+ @match_cases) @match_statement
    """,
    "case_clause": """
    (case_clause
        pattern: (_) @case_pattern
        guard: (_)? @case_guard
        consequence: (block) @case_body) @case_clause
    """,
    "walrus_operator": """
    (named_expression
        name: (identifier) @walrus_target
        value: (_) @walrus_value) @walrus_operator
    """,
    "f_string": """
    (formatted_string
        (interpolation) @f_string_interpolation) @f_string
    """,
    "yield_expression": """
    (yield
        (_)? @yielded_value) @yield_expression
    """,
    "yield_from": """
    (yield
        "from" @yield_from_keyword
        (_) @yielded_iterable) @yield_from_expression
    """,
    "await_expression": """
    (await
        (_) @awaited_expression) @await_expression
    """,
    # --- Comments and Docstrings ---
    "comment": """
    (comment) @comment
    """,
    "docstring": """
    (expression_statement
        (string) @docstring)
    """,
    "module_docstring": """
    (module
        (expression_statement
            (string) @module_docstring))
    """,
    # --- Framework-specific Patterns ---
    "django_model": """
    (class_definition
        name: (identifier) @model_name
        superclasses: (argument_list
            (identifier) @superclass
            (#match? @superclass "Model|models.Model"))
        body: (block) @model_body) @django_model
    """,
    "django_view": """
    (class_definition
        name: (identifier) @view_name
        superclasses: (argument_list
            (identifier) @superclass
            (#match? @superclass "View|TemplateView|ListView|DetailView"))
        body: (block) @view_body) @django_view
    """,
    "flask_route": """
    (decorated_definition
        (decorator
            (call
                function: (attribute
                    object: (identifier) @app_object
                    attribute: (identifier) @route_decorator
                    (#match? @route_decorator "route"))
                arguments: (argument_list
                    (string) @route_path))) @flask_route_decorator
        (function_definition
            name: (identifier) @handler_name)) @flask_route
    """,
    "fastapi_endpoint": """
    (decorated_definition
        (decorator
            (call
                function: (attribute
                    object: (identifier) @app_object
                    attribute: (identifier) @http_method
                    (#match? @http_method "get|post|put|delete|patch"))
                arguments: (argument_list
                    (string) @endpoint_path))) @fastapi_decorator
        (function_definition
            name: (identifier) @endpoint_name)) @fastapi_endpoint
    """,
    "dataclass": """
    (decorated_definition
        (decorator
            (identifier) @decorator_name
            (#match? @decorator_name "dataclass"))
        (class_definition
            name: (identifier) @dataclass_name)) @dataclass_definition
    """,
    # --- Name-only Extraction ---
    "function_name": """
    (function_definition
        name: (identifier) @function_name)
    """,
    "class_name": """
    (class_definition
        name: (identifier) @class_name)
    """,
    "variable_name": """
    (assignment
        left: (identifier) @variable_name)
    """,
    # --- Advanced Patterns ---
    "context_manager": """
    (class_definition
        body: (block
            (function_definition
                name: (identifier) @enter_method
                (#match? @enter_method "__enter__"))
            (function_definition
                name: (identifier) @exit_method
                (#match? @exit_method "__exit__")))) @context_manager_class
    """,
    "iterator": """
    (class_definition
        body: (block
            (function_definition
                name: (identifier) @iter_method
                (#match? @iter_method "__iter__"))
            (function_definition
                name: (identifier) @next_method
                (#match? @next_method "__next__")))) @iterator_class
    """,
    "metaclass": """
    (class_definition
        name: (identifier) @metaclass_name
        superclasses: (argument_list
            (identifier) @superclass
            (#match? @superclass "type"))
        body: (block) @metaclass_body) @metaclass_definition
    """,
    "abstract_method": """
    (decorated_definition
        (decorator
            (identifier) @decorator_name
            (#match? @decorator_name "abstractmethod"))
        (function_definition
            name: (identifier) @abstract_method_name)) @abstract_method
    """,
}

# Query descriptions
PYTHON_QUERY_DESCRIPTIONS: dict[str, str] = {
    "function": "Search Python function definitions",
    "function_definition": "Search function definitions with details",
    "async_function": "Search async function definitions",
    "method": "Search method definitions within classes",
    "lambda": "Search lambda expressions",
    "class": "Search Python class definitions",
    "class_definition": "Search class definitions with details",
    "class_method": "Search class methods",
    "constructor": "Search class constructors (__init__)",
    "property": "Search property decorators",
    "staticmethod": "Search static methods",
    "classmethod": "Search class methods",
    "variable": "Search variable assignments",
    "assignment": "Search variable assignments",
    "multiple_assignment": "Search multiple variable assignments",
    "augmented_assignment": "Search augmented assignments (+=, -=, etc.)",
    "global_statement": "Search global variable declarations",
    "nonlocal_statement": "Search nonlocal variable declarations",
    "import": "Search import statements",
    "import_statement": "Search import statements with details",
    "import_from": "Search from-import statements",
    "import_from_list": "Search from-import with multiple names",
    "aliased_import": "Search aliased imports (as keyword)",
    "import_star": "Search star imports (from module import *)",
    "decorator": "Search all decorators",
    "decorator_simple": "Search simple decorators",
    "decorator_call": "Search decorator calls with arguments",
    "decorator_attribute": "Search attribute decorators",
    "decorated_function": "Search decorated functions",
    "decorated_class": "Search decorated classes",
    "if_statement": "Search if statements",
    "for_statement": "Search for loops",
    "while_statement": "Search while loops",
    "with_statement": "Search with statements (context managers)",
    "async_with": "Search async with statements",
    "async_for": "Search async for loops",
    "try_statement": "Search try-except statements",
    "except_clause": "Search except clauses",
    "raise_statement": "Search raise statements",
    "assert_statement": "Search assert statements",
    "list_comprehension": "Search list comprehensions",
    "dict_comprehension": "Search dictionary comprehensions",
    "set_comprehension": "Search set comprehensions",
    "generator_expression": "Search generator expressions",
    "type_hint": "Search type hints on parameters",
    "return_type": "Search function return type annotations",
    "variable_annotation": "Search variable type annotations",
    "type_alias": "Search type alias statements",
    "match_statement": "Search match statements (Python 3.10+)",
    "case_clause": "Search case clauses in match statements",
    "walrus_operator": "Search walrus operator (:=)",
    "f_string": "Search f-string literals",
    "yield_expression": "Search yield expressions",
    "yield_from": "Search yield from expressions",
    "await_expression": "Search await expressions",
    "comment": "Search all comments",
    "docstring": "Search docstrings",
    "module_docstring": "Search module-level docstrings",
    "django_model": "Search Django model classes",
    "django_view": "Search Django view classes",
    "flask_route": "Search Flask route decorators",
    "fastapi_endpoint": "Search FastAPI endpoint decorators",
    "dataclass": "Search dataclass definitions",
    "function_name": "Search function names only",
    "class_name": "Search class names only",
    "variable_name": "Search variable names only",
    "context_manager": "Search context manager classes",
    "iterator": "Search iterator classes",
    "metaclass": "Search metaclass definitions",
    "abstract_method": "Search abstract methods",
}

# Convert to ALL_QUERIES format for dynamic loader compatibility
ALL_QUERIES = {}
for query_name, query_string in PYTHON_QUERIES.items():
    description = PYTHON_QUERY_DESCRIPTIONS.get(query_name, "No description")
    ALL_QUERIES[query_name] = {"query": query_string, "description": description}

# Add legacy queries for backward compatibility
ALL_QUERIES["functions"] = {
    "query": FUNCTIONS,
    "description": "Search all function definitions (including async)",
}
ALL_QUERIES["classes"] = {
    "query": CLASSES,
    "description": "Search all class definitions",
}
ALL_QUERIES["imports"] = {
    "query": IMPORTS,
    "description": "Search all import statements",
}
ALL_QUERIES["variables"] = {
    "query": VARIABLES,
    "description": "Search all variable assignments",
}
ALL_QUERIES["decorators"] = {
    "query": DECORATORS,
    "description": "Search all decorators",
}
ALL_QUERIES["methods"] = {
    "query": METHODS,
    "description": "Search all method definitions within classes",
}
ALL_QUERIES["exceptions"] = {
    "query": EXCEPTIONS,
    "description": "Search exception handling and raise statements",
}
ALL_QUERIES["comprehensions"] = {
    "query": COMPREHENSIONS,
    "description": "Search list, dictionary, and set comprehensions",
}
ALL_QUERIES["comments"] = {
    "query": COMMENTS,
    "description": "Search comments and docstrings",
}
ALL_QUERIES["type_hints"] = {
    "query": TYPE_HINTS,
    "description": "Search type hints and annotations",
}
ALL_QUERIES["async_patterns"] = {
    "query": ASYNC_PATTERNS,
    "description": "Search async/await patterns",
}
ALL_QUERIES["string_formatting"] = {
    "query": STRING_FORMATTING,
    "description": "Search f-strings and string formatting",
}
ALL_QUERIES["context_managers"] = {
    "query": CONTEXT_MANAGERS,
    "description": "Search context managers (with statements)",
}
ALL_QUERIES["lambdas"] = {"query": LAMBDAS, "description": "Search lambda expressions"}
ALL_QUERIES["modern_patterns"] = {
    "query": MODERN_PATTERNS,
    "description": "Search modern Python patterns (match/case, walrus operator)",
}

# Convenience aliases
ALL_QUERIES["function_names"] = {
    "query": FUNCTIONS,
    "description": "Function alias - search all function definitions",
}
ALL_QUERIES["class_names"] = {
    "query": CLASSES,
    "description": "Class alias - search all class definitions",
}
ALL_QUERIES["all_declarations"] = {
    "query": FUNCTIONS + "\n\n" + CLASSES + "\n\n" + VARIABLES,
    "description": "Search all function, class, and variable declarations",
}


def get_python_query(name: str) -> str:
    """
    Get the specified Python query

    Args:
        name: Query name

    Returns:
        Query string

    Raises:
        ValueError: When query is not found
    """
    if name not in PYTHON_QUERIES:
        available = list(PYTHON_QUERIES.keys())
        raise ValueError(
            f"Python query '{name}' does not exist. Available: {available}"
        )

    return PYTHON_QUERIES[name]


def get_python_query_description(name: str) -> str:
    """
    Get the description of the specified Python query

    Args:
        name: Query name

    Returns:
        Query description
    """
    return PYTHON_QUERY_DESCRIPTIONS.get(name, "No description")


def get_query(name: str) -> str:
    """Get a specific query by name."""
    if name in ALL_QUERIES:
        return ALL_QUERIES[name]["query"]
    raise ValueError(
        f"Query '{name}' not found. Available queries: {list(ALL_QUERIES.keys())}"
    )


def get_all_queries() -> dict:
    """Get all available queries."""
    return ALL_QUERIES


def list_queries() -> list:
    """List all available query names."""
    return list(ALL_QUERIES.keys())


def get_available_python_queries() -> list[str]:
    """
    Get list of available Python queries

    Returns:
        List of query names
    """
    return list(PYTHON_QUERIES.keys())



#!/usr/bin/env python3
"""
Project Root Detection

Intelligent detection of project root directories based on common project markers.
"""

import logging
from pathlib import Path

logger = logging.getLogger(__name__)

# Common project root indicators (in priority order)
PROJECT_MARKERS = [
    # Version control
    ".git",
    ".hg",
    ".svn",
    # Python projects
    "pyproject.toml",
    "setup.py",
    "setup.cfg",
    "requirements.txt",
    "Pipfile",
    "poetry.lock",
    "conda.yaml",
    "environment.yml",
    # JavaScript/Node.js projects
    "package.json",
    "package-lock.json",
    "yarn.lock",
    "node_modules",
    # Java projects
    "pom.xml",
    "build.gradle",
    "build.gradle.kts",
    "gradlew",
    "mvnw",
    # C/C++ projects
    "CMakeLists.txt",
    "Makefile",
    "configure.ac",
    "configure.in",
    # Rust projects
    "Cargo.toml",
    "Cargo.lock",
    # Go projects
    "go.mod",
    "go.sum",
    # .NET projects
    "*.sln",
    "*.csproj",
    "*.vbproj",
    "*.fsproj",
    # Other common markers
    "README.md",
    "README.rst",
    "README.txt",
    "LICENSE",
    "CHANGELOG.md",
    ".dockerignore",
    "Dockerfile",
    "docker-compose.yml",
    ".editorconfig",
]


class ProjectRootDetector:
    """Intelligent project root directory detection."""

    def __init__(self, max_depth: int = 10):
        """
        Initialize project root detector.

        Args:
            max_depth: Maximum directory levels to traverse upward
        """
        self.max_depth = max_depth

    def detect_from_file(self, file_path: str) -> str | None:
        """
        Detect project root from a file path.

        Args:
            file_path: Path to a file within the project

        Returns:
            Project root directory path, or None if not detected
        """
        if not file_path:
            return None

        try:
            # Convert to absolute path and get directory
            abs_path = Path(file_path).resolve()
            if abs_path.is_file():
                start_dir = abs_path.parent
            else:
                start_dir = abs_path

            return self._traverse_upward(str(start_dir))

        except Exception as e:
            logger.warning(f"Error detecting project root from {file_path}: {e}")
            return None

    def detect_from_cwd(self) -> str | None:
        """
        Detect project root from current working directory.

        Returns:
            Project root directory path, or None if not detected
        """
        try:
            return self._traverse_upward(str(Path.cwd()))
        except Exception as e:
            logger.warning(f"Error detecting project root from cwd: {e}")
            return None

    def _traverse_upward(self, start_dir: str) -> str | None:
        """
        Traverse upward from start directory looking for project markers.

        Args:
            start_dir: Directory to start traversal from

        Returns:
            Project root directory path, or None if not found
        """
        current_dir = str(Path(start_dir).resolve())
        candidates = []

        for _depth in range(self.max_depth):
            # Check for project markers in current directory
            markers_found = self._find_markers_in_dir(current_dir)

            if markers_found:
                # Calculate score based on marker priority and count
                score = self._calculate_score(markers_found)
                candidates.append((current_dir, score, markers_found))

                # If we find high-priority markers, we can stop early
                if any(
                    marker
                    in [
                        ".git",
                        "pyproject.toml",
                        "package.json",
                        "pom.xml",
                        "Cargo.toml",
                        "go.mod",
                    ]
                    for marker in markers_found
                ):
                    logger.debug(
                        f"Found high-priority project root: {current_dir} (markers: {markers_found})"
                    )
                    return current_dir

            # Move up one directory
            current_path = Path(current_dir)
            parent_path = current_path.parent
            if parent_path == current_path:  # Reached filesystem root
                break
            current_dir = str(parent_path)

        # Return the best candidate if any found
        if candidates:
            # Sort by score (descending) and return the best
            candidates.sort(key=lambda x: x[1], reverse=True)
            best_candidate = candidates[0]
            logger.debug(
                f"Selected project root: {best_candidate[0]} (score: {best_candidate[1]}, markers: {best_candidate[2]})"
            )
            return best_candidate[0]

        logger.debug(f"No project root detected from {start_dir}")
        return None

    def _find_markers_in_dir(self, directory: str) -> list[str]:
        """
        Find project markers in a directory.

        Args:
            directory: Directory to search in

        Returns:
            List of found marker names
        """
        found_markers = []

        try:
            dir_path = Path(directory)

            for marker in PROJECT_MARKERS:
                if "*" in marker:
                    # Handle glob patterns using pathlib
                    if list(dir_path.glob(marker)):
                        found_markers.append(marker)
                else:
                    # Handle exact matches
                    if (dir_path / marker).exists():
                        found_markers.append(marker)

        except (OSError, PermissionError) as e:
            logger.debug(f"Cannot access directory {directory}: {e}")

        return found_markers

    def _calculate_score(self, markers: list[str]) -> int:
        """
        Calculate a score for project root candidates based on markers found.

        Args:
            markers: List of found markers

        Returns:
            Score (higher is better)
        """
        score = 0

        # High-priority markers
        high_priority = [
            ".git",
            "pyproject.toml",
            "package.json",
            "pom.xml",
            "Cargo.toml",
            "go.mod",
        ]
        medium_priority = ["setup.py", "requirements.txt", "CMakeLists.txt", "Makefile"]

        for marker in markers:
            if marker in high_priority:
                score += 100
            elif marker in medium_priority:
                score += 50
            else:
                score += 10

        # Bonus for multiple markers
        if len(markers) > 1:
            score += len(markers) * 5

        return score

    def get_fallback_root(self, file_path: str) -> str:
        """
        Get fallback project root when detection fails.

        Args:
            file_path: Original file path

        Returns:
            Fallback directory (file's directory or cwd)
        """
        try:
            if file_path:
                path = Path(file_path)
                if path.exists():
                    if path.is_file():
                        return str(path.resolve().parent)
                    else:
                        return str(path.resolve())
            return str(Path.cwd())
        except Exception:
            return str(Path.cwd())


def detect_project_root(
    file_path: str | None = None, explicit_root: str | None = None
) -> str | None:
    """
    Unified project root detection with priority handling.

    Priority order:
    1. explicit_root parameter (highest priority)
    2. Auto-detection from file_path
    3. Auto-detection from current working directory
    4. Return None if no markers found

    Args:
        file_path: Path to a file within the project
        explicit_root: Explicitly specified project root

    Returns:
        Project root directory path, or None if no markers found
    """
    detector = ProjectRootDetector()

    # Priority 1: Explicit root
    if explicit_root:
        explicit_path = Path(explicit_root)
        if explicit_path.exists() and explicit_path.is_dir():
            logger.debug(f"Using explicit project root: {explicit_root}")
            return str(explicit_path.resolve())
        else:
            logger.warning(f"Explicit project root does not exist: {explicit_root}")

    # Priority 2: Auto-detection from file path
    if file_path:
        detected_root = detector.detect_from_file(file_path)
        if detected_root:
            logger.debug(f"Auto-detected project root from file: {detected_root}")
            return detected_root

    # Priority 3: Auto-detection from cwd
    detected_root = detector.detect_from_cwd()
    if detected_root:
        logger.debug(f"Auto-detected project root from cwd: {detected_root}")
        return detected_root

    # Priority 4: Return None if no markers found
    logger.debug("No project markers found, returning None")
    return None


if __name__ == "__main__":
    # Test the detector
    import sys

    if len(sys.argv) > 1:
        test_path = sys.argv[1]
        result = detect_project_root(test_path)
        print(f"Project root for '{test_path}': {result}")
    else:
        result = detect_project_root()
        print(f"Project root from cwd: {result}")



#!/usr/bin/env python3
"""
YAML Language Plugin

YAML-specific parsing and element extraction functionality using tree-sitter-yaml.
Provides comprehensive support for YAML elements including mappings, sequences,
scalars, anchors, aliases, and comments.
"""

import logging
from typing import TYPE_CHECKING, Any

from ..models import AnalysisResult, CodeElement
from ..plugins.base import ElementExtractor, LanguagePlugin
from ..utils import log_debug, log_error, log_info, log_warning

if TYPE_CHECKING:
    import tree_sitter

    from ..core.analysis_engine import AnalysisRequest

logger = logging.getLogger(__name__)

# Graceful degradation for tree-sitter-yaml
try:
    import tree_sitter
    import tree_sitter_yaml as ts_yaml

    YAML_AVAILABLE = True
except ImportError:
    YAML_AVAILABLE = False
    log_warning("tree-sitter-yaml not installed, YAML support disabled")


class YAMLElement(CodeElement):
    """YAML-specific code element."""

    def __init__(
        self,
        name: str,
        start_line: int,
        end_line: int,
        raw_text: str,
        language: str = "yaml",
        element_type: str = "yaml",
        key: str | None = None,
        value: str | None = None,
        value_type: str | None = None,
        anchor_name: str | None = None,
        alias_target: str | None = None,
        nesting_level: int = 0,
        document_index: int = 0,
        child_count: int | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize YAMLElement.

        Args:
            name: Element name
            start_line: Starting line number
            end_line: Ending line number
            raw_text: Raw text content
            language: Language identifier
            element_type: Type of YAML element
            key: Key for mapping pairs
            value: Scalar value (None for complex structures)
            value_type: Type of value (string, number, boolean, null, mapping, sequence)
            anchor_name: Anchor name for &anchor definitions
            alias_target: Target anchor name for *alias references (not resolved)
            nesting_level: AST-based logical depth
            document_index: Index of document in multi-document YAML
            child_count: Number of child elements for complex structures
            **kwargs: Additional attributes
        """
        super().__init__(
            name=name,
            start_line=start_line,
            end_line=end_line,
            raw_text=raw_text,
            language=language,
            **kwargs,
        )
        self.element_type = element_type
        self.key = key
        self.value = value
        self.value_type = value_type
        self.anchor_name = anchor_name
        self.alias_target = alias_target
        self.nesting_level = nesting_level
        self.document_index = document_index
        self.child_count = child_count


class YAMLElementExtractor(ElementExtractor):
    """YAML-specific element extractor using tree-sitter-yaml."""

    def __init__(self) -> None:
        """Initialize the YAML element extractor."""
        self.source_code: str = ""
        self.content_lines: list[str] = []
        self._current_document_index: int = 0

    def extract_functions(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[CodeElement]:
        """YAML doesn't have functions, return empty list."""
        return []

    def extract_classes(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[CodeElement]:
        """YAML doesn't have classes, return empty list."""
        return []

    def extract_variables(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[CodeElement]:
        """YAML doesn't have variables, return empty list."""
        return []

    def extract_imports(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[CodeElement]:
        """YAML doesn't have imports, return empty list."""
        return []

    def extract_yaml_elements(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[YAMLElement]:
        """Extract all YAML elements from the parsed tree.

        Args:
            tree: Parsed tree-sitter tree
            source_code: Original source code

        Returns:
            List of YAMLElement objects
        """
        self.source_code = source_code or ""
        self.content_lines = self.source_code.split("\n")
        self._current_document_index = 0

        elements: list[YAMLElement] = []

        if tree is None or tree.root_node is None:
            return elements

        try:
            # Extract documents first to set document indices
            self._extract_documents(tree.root_node, elements)
            # Extract mappings
            self._extract_mappings(tree.root_node, elements)
            # Extract sequences
            self._extract_sequences(tree.root_node, elements)
            # Extract anchors and aliases
            self._extract_anchors(tree.root_node, elements)
            self._extract_aliases(tree.root_node, elements)
            # Extract comments
            self._extract_comments(tree.root_node, elements)
        except Exception as e:
            log_error(f"Error during YAML element extraction: {e}")

        log_debug(f"Extracted {len(elements)} YAML elements")
        return elements

    def _get_node_text(self, node: "tree_sitter.Node") -> str:
        """Get text content from a tree-sitter node."""
        try:
            if hasattr(node, "start_byte") and hasattr(node, "end_byte"):
                source_bytes = self.source_code.encode("utf-8")
                node_bytes = source_bytes[node.start_byte : node.end_byte]
                return node_bytes.decode("utf-8", errors="replace")
            return ""
        except Exception as e:
            log_debug(f"Failed to extract node text: {e}")
            return ""

    def _calculate_nesting_level(self, node: "tree_sitter.Node") -> int:
        """Calculate AST-based logical nesting level."""
        level = 0
        current = node.parent
        while current is not None:
            if current.type in (
                "block_mapping",
                "block_sequence",
                "flow_mapping",
                "flow_sequence",
            ):
                level += 1
            current = current.parent
        return level

    def _get_document_index(self, node: "tree_sitter.Node") -> int:
        """Get document index for a node."""
        current = node
        while current is not None:
            if current.type == "document":
                # Count preceding document siblings
                index = 0
                sibling = current.prev_sibling
                while sibling is not None:
                    if sibling.type == "document":
                        index += 1
                    sibling = sibling.prev_sibling
                return index
            current = current.parent
        return 0

    def _traverse_nodes(self, node: "tree_sitter.Node") -> "list[tree_sitter.Node]":
        """Traverse all nodes in the tree."""
        nodes = [node]
        for child in node.children:
            nodes.extend(self._traverse_nodes(child))
        return nodes

    def _count_document_children(self, document_node: "tree_sitter.Node") -> int:
        """Count meaningful children in a document (top-level mappings).

        This counts the number of top-level key-value pairs in the document,
        which is more meaningful than counting AST nodes.
        """
        count = 0
        for child in document_node.children:
            # Skip document markers and comments
            if child.type in ("---", "...", "comment"):
                continue
            # For block_node, count the mappings inside
            if child.type == "block_node":
                for subchild in child.children:
                    if subchild.type == "block_mapping":
                        # Count the mapping pairs
                        count += len(
                            [
                                c
                                for c in subchild.children
                                if c.type == "block_mapping_pair"
                            ]
                        )
                    elif subchild.type in ("block_sequence", "flow_sequence"):
                        count += 1
            elif child.type == "block_mapping":
                count += len(
                    [c for c in child.children if c.type == "block_mapping_pair"]
                )
        return count

    def _extract_documents(
        self, root_node: "tree_sitter.Node", elements: list[YAMLElement]
    ) -> None:
        """Extract YAML documents."""
        for node in self._traverse_nodes(root_node):
            if node.type == "document":
                try:
                    start_line = node.start_point[0] + 1
                    end_line = node.end_point[0] + 1
                    raw_text = self._get_node_text(node)
                    doc_index = self._get_document_index(node)

                    # Count meaningful child elements (top-level mappings)
                    # Exclude document markers (---) and comments
                    child_count = self._count_document_children(node)

                    element = YAMLElement(
                        name=f"Document {doc_index}",
                        start_line=start_line,
                        end_line=end_line,
                        raw_text=raw_text[:200] + "..."
                        if len(raw_text) > 200
                        else raw_text,
                        element_type="document",
                        document_index=doc_index,
                        child_count=child_count,
                        nesting_level=0,
                    )
                    elements.append(element)
                except Exception as e:
                    log_debug(f"Failed to extract document: {e}")

    def _extract_mappings(
        self, root_node: "tree_sitter.Node", elements: list[YAMLElement]
    ) -> None:
        """Extract YAML mappings (key-value pairs)."""
        for node in self._traverse_nodes(root_node):
            if node.type in ("block_mapping_pair", "flow_pair"):
                try:
                    start_line = node.start_point[0] + 1
                    end_line = node.end_point[0] + 1
                    raw_text = self._get_node_text(node)

                    # Extract key and value
                    key = None
                    value = None
                    value_type = None
                    child_count = None

                    # Find key and value nodes
                    # In tree-sitter-yaml, block_mapping_pair has structure:
                    # flow_node (key), ':', flow_node (value)
                    key_node = None
                    value_node = None
                    found_colon = False

                    for child in node.children:
                        if child.type == ":":
                            found_colon = True
                        elif child.type in ("flow_node", "block_node"):
                            if not found_colon:
                                # This is the key
                                key_node = child
                            else:
                                # This is the value
                                value_node = child
                        elif child.type == "key":
                            # Key is wrapped in a "key" node
                            if child.children:
                                key_node = child.children[0]
                            else:
                                key_node = child
                        elif child.type == "value":
                            # Value is wrapped in a "value" node
                            if child.children:
                                value_node = child.children[0]
                            else:
                                value_node = child

                    # Extract key text - drill down through flow_node/block_node
                    if key_node is not None:
                        # Drill down to get the actual scalar
                        current = key_node
                        while (
                            current
                            and current.type in ("flow_node", "block_node")
                            and current.children
                        ):
                            current = current.children[0]
                        if current:
                            key = self._get_node_text(current).strip()

                    # Extract value info - drill down through flow_node/block_node
                    if value_node is not None:
                        # Drill down to get the actual value node
                        current = value_node
                        while (
                            current
                            and current.type in ("flow_node", "block_node")
                            and current.children
                        ):
                            current = current.children[0]
                        if current:
                            value, value_type, child_count = self._extract_value_info(
                                current
                            )

                    nesting_level = self._calculate_nesting_level(node)
                    doc_index = self._get_document_index(node)

                    element = YAMLElement(
                        name=key or "mapping",
                        start_line=start_line,
                        end_line=end_line,
                        raw_text=raw_text,
                        element_type="mapping",
                        key=key,
                        value=value,
                        value_type=value_type,
                        nesting_level=nesting_level,
                        document_index=doc_index,
                        child_count=child_count,
                    )
                    elements.append(element)
                except Exception as e:
                    log_debug(f"Failed to extract mapping: {e}")

    def _extract_value_info(
        self, node: "tree_sitter.Node"
    ) -> tuple[str | None, str | None, int | None]:
        """Extract value information from a node.

        Returns:
            Tuple of (value, value_type, child_count)
        """
        if node is None:
            return None, None, None

        node_type = node.type
        text = self._get_node_text(node).strip()

        # Scalar types
        if node_type in ("plain_scalar", "double_quote_scalar", "single_quote_scalar"):
            # Determine scalar type
            if text.lower() in ("true", "false", "yes", "no", "on", "off"):
                return text, "boolean", None
            elif text.lower() in ("null", "~", ""):
                return text if text else None, "null", None
            elif self._is_number(text):
                return text, "number", None
            else:
                return text, "string", None
        elif node_type == "block_scalar":
            return text, "string", None
        elif node_type in ("block_mapping", "flow_mapping"):
            child_count = len(
                [
                    c
                    for c in node.children
                    if c.type in ("block_mapping_pair", "flow_pair")
                ]
            )
            return None, "mapping", child_count
        elif node_type in ("block_sequence", "flow_sequence"):
            child_count = len(
                [c for c in node.children if c.type in ("block_sequence_item",)]
                or node.children
            )
            return None, "sequence", child_count
        elif node_type == "alias":
            alias_name = text.lstrip("*")
            return f"*{alias_name}", "alias", None

        return text, "unknown", None

    def _is_number(self, text: str) -> bool:
        """Check if text represents a number."""
        try:
            float(text)
            return True
        except ValueError:
            return False

    def _extract_sequences(
        self, root_node: "tree_sitter.Node", elements: list[YAMLElement]
    ) -> None:
        """Extract YAML sequences (lists)."""
        for node in self._traverse_nodes(root_node):
            if node.type in ("block_sequence", "flow_sequence"):
                try:
                    start_line = node.start_point[0] + 1
                    end_line = node.end_point[0] + 1
                    raw_text = self._get_node_text(node)

                    # Count items
                    if node.type == "block_sequence":
                        child_count = len(
                            [
                                c
                                for c in node.children
                                if c.type == "block_sequence_item"
                            ]
                        )
                    else:
                        child_count = len(node.children)

                    nesting_level = self._calculate_nesting_level(node)
                    doc_index = self._get_document_index(node)

                    element = YAMLElement(
                        name="sequence",
                        start_line=start_line,
                        end_line=end_line,
                        raw_text=raw_text[:200] + "..."
                        if len(raw_text) > 200
                        else raw_text,
                        element_type="sequence",
                        value_type="sequence",
                        nesting_level=nesting_level,
                        document_index=doc_index,
                        child_count=child_count,
                    )
                    elements.append(element)
                except Exception as e:
                    log_debug(f"Failed to extract sequence: {e}")

    def _extract_anchors(
        self, root_node: "tree_sitter.Node", elements: list[YAMLElement]
    ) -> None:
        """Extract YAML anchors (&name)."""
        for node in self._traverse_nodes(root_node):
            if node.type == "anchor":
                try:
                    start_line = node.start_point[0] + 1
                    end_line = node.end_point[0] + 1
                    raw_text = self._get_node_text(node)
                    anchor_name = raw_text.lstrip("&").strip()

                    nesting_level = self._calculate_nesting_level(node)
                    doc_index = self._get_document_index(node)

                    element = YAMLElement(
                        name=f"&{anchor_name}",
                        start_line=start_line,
                        end_line=end_line,
                        raw_text=raw_text,
                        element_type="anchor",
                        anchor_name=anchor_name,
                        nesting_level=nesting_level,
                        document_index=doc_index,
                    )
                    elements.append(element)
                except Exception as e:
                    log_debug(f"Failed to extract anchor: {e}")

    def _extract_aliases(
        self, root_node: "tree_sitter.Node", elements: list[YAMLElement]
    ) -> None:
        """Extract YAML aliases (*name)."""
        for node in self._traverse_nodes(root_node):
            if node.type == "alias":
                try:
                    start_line = node.start_point[0] + 1
                    end_line = node.end_point[0] + 1
                    raw_text = self._get_node_text(node)
                    alias_target = raw_text.lstrip("*").strip()

                    nesting_level = self._calculate_nesting_level(node)
                    doc_index = self._get_document_index(node)

                    element = YAMLElement(
                        name=f"*{alias_target}",
                        start_line=start_line,
                        end_line=end_line,
                        raw_text=raw_text,
                        element_type="alias",
                        alias_target=alias_target,
                        nesting_level=nesting_level,
                        document_index=doc_index,
                    )
                    elements.append(element)
                except Exception as e:
                    log_debug(f"Failed to extract alias: {e}")

    def _extract_comments(
        self, root_node: "tree_sitter.Node", elements: list[YAMLElement]
    ) -> None:
        """Extract YAML comments."""
        for node in self._traverse_nodes(root_node):
            if node.type == "comment":
                try:
                    start_line = node.start_point[0] + 1
                    end_line = node.end_point[0] + 1
                    raw_text = self._get_node_text(node)
                    comment_text = raw_text.lstrip("#").strip()

                    doc_index = self._get_document_index(node)

                    element = YAMLElement(
                        name=comment_text[:50] + "..."
                        if len(comment_text) > 50
                        else comment_text,
                        start_line=start_line,
                        end_line=end_line,
                        raw_text=raw_text,
                        element_type="comment",
                        value=comment_text,
                        value_type="comment",
                        document_index=doc_index,
                        nesting_level=0,
                    )
                    elements.append(element)
                except Exception as e:
                    log_debug(f"Failed to extract comment: {e}")


class YAMLPlugin(LanguagePlugin):
    """YAML language plugin using tree-sitter-yaml for true YAML parsing."""

    def get_language_name(self) -> str:
        """Return the language name."""
        return "yaml"

    def get_file_extensions(self) -> list[str]:
        """Return supported file extensions."""
        return [".yaml", ".yml"]

    def create_extractor(self) -> ElementExtractor:
        """Create and return a YAML element extractor."""
        return YAMLElementExtractor()

    def get_supported_element_types(self) -> list[str]:
        """Return supported element types."""
        return [
            "mapping",
            "sequence",
            "scalar",
            "anchor",
            "alias",
            "comment",
            "document",
        ]

    def get_queries(self) -> dict[str, str]:
        """Return YAML-specific tree-sitter queries."""
        from ..queries.yaml import YAML_QUERIES

        return YAML_QUERIES

    def execute_query_strategy(
        self, query_key: str | None, language: str
    ) -> str | None:
        """Execute query strategy for YAML."""
        if language != "yaml":
            return None

        queries = self.get_queries()
        return queries.get(query_key) if query_key else None

    def get_element_categories(self) -> dict[str, list[str]]:
        """Return YAML element categories for query execution."""
        return {
            "structure": ["document", "block_mapping", "block_sequence"],
            "mappings": ["block_mapping_pair", "flow_pair"],
            "sequences": ["block_sequence", "flow_sequence"],
            "scalars": [
                "plain_scalar",
                "double_quote_scalar",
                "single_quote_scalar",
                "block_scalar",
            ],
            "references": ["anchor", "alias"],
            "metadata": ["comment", "tag"],
        }

    async def analyze_file(
        self, file_path: str, request: "AnalysisRequest"
    ) -> "AnalysisResult":
        """Analyze YAML file using tree-sitter-yaml parser.

        Args:
            file_path: Path to the YAML file
            request: Analysis request parameters

        Returns:
            AnalysisResult with extracted elements
        """
        from ..encoding_utils import read_file_safe

        # Check if YAML support is available
        if not YAML_AVAILABLE:
            log_error("tree-sitter-yaml not available")
            return AnalysisResult(
                file_path=file_path,
                language="yaml",
                line_count=0,
                elements=[],
                node_count=0,
                query_results={},
                source_code="",
                success=False,
                error_message="YAML support not available. Install tree-sitter-yaml.",
            )

        try:
            # Read file content with encoding detection
            content, encoding = read_file_safe(file_path)

            # Get YAML language
            YAML_LANGUAGE = tree_sitter.Language(ts_yaml.language())

            # Create parser
            parser = tree_sitter.Parser()
            parser.language = YAML_LANGUAGE

            # Parse the YAML content
            tree = parser.parse(content.encode("utf-8"))

            # Extract elements using the extractor
            extractor = self.create_extractor()
            elements = extractor.extract_yaml_elements(tree, content)

            log_info(f"Extracted {len(elements)} YAML elements from {file_path}")

            return AnalysisResult(
                file_path=file_path,
                language="yaml",
                line_count=len(content.splitlines()),
                elements=elements,
                node_count=len(elements),
                query_results={},
                source_code=content,
                success=True,
                error_message=None,
            )

        except Exception as e:
            log_error(f"Failed to analyze YAML file {file_path}: {e}")
            return AnalysisResult(
                file_path=file_path,
                language="yaml",
                line_count=0,
                elements=[],
                node_count=0,
                query_results={},
                source_code="",
                success=False,
                error_message=str(e),
            )



#!/usr/bin/env python3
"""
Smart Cache Demo - Cross-Format Optimization

This demo showcases the intelligent cross-format caching optimization where:
1. First search gets match counts (fast --count-matches)
2. Second search requests file list but gets it derived from cached count data
3. No second ripgrep execution needed!

Performance improvement: From 2 searches → 1 search + instant derivation
"""

import asyncio
import time
from pathlib import Path

from tree_sitter_analyzer.mcp.tools.search_content_tool import SearchContentTool
from tree_sitter_analyzer.mcp.utils.search_cache import (
    clear_cache,
    configure_cache,
    get_default_cache,
)


async def demo_smart_cache_optimization():
    print("🧠 Smart Cache Cross-Format Optimization Demo")
    print("=" * 60)

    # Use the project's examples directory for testing
    script_dir = Path(__file__).parent
    test_directory = str(script_dir)

    if not Path(test_directory).exists():
        print(f"❌ Test directory not found: {test_directory}")
        return

    print(f"📁 Using test directory: {test_directory}")
    print()

    # Configure cache
    configure_cache(max_size=100, ttl_seconds=300)
    clear_cache()  # Ensure clean slate

    # Initialize tool with caching enabled
    tool = SearchContentTool(project_root=test_directory, enable_cache=True)

    print("🔍 Scenario: LLM workflow optimization")
    print("   Step 1: Get match count (for token efficiency)")
    print("   Step 2: Get file list (for detailed analysis)")
    print()

    # Step 1: Get match count (typical LLM workflow)
    print("📊 Step 1: Getting match count...")
    start_time = time.time()

    count_result = await tool.execute(
        {"query": "class", "roots": [test_directory], "count_only_matches": True}
    )

    step1_time = time.time() - start_time
    print(f"   ⏱️  Time: {step1_time:.3f}s")
    print(f"   📈 Total matches: {count_result.get('total_matches', 0)}")
    print(f"   📁 Files with matches: {len(count_result.get('file_counts', {}))}")

    # Show cache stats
    cache_stats = get_default_cache().get_stats()
    print(
        f"   🎯 Cache stats: {cache_stats['hits']} hits, {cache_stats['misses']} misses"
    )
    print()

    # Step 2: Get file summary (should derive from cached count data!)
    print("📋 Step 2: Getting file summary (watch for cache derivation)...")
    start_time = time.time()

    summary_result = await tool.execute(
        {"query": "class", "roots": [test_directory], "summary_only": True}
    )

    step2_time = time.time() - start_time
    print(f"   ⏱️  Time: {step2_time:.3f}s")

    # Check if result was derived from cache
    if summary_result.get("cache_derived"):
        print("   🎉 SUCCESS: File list derived from cached count data!")
        print("   🚀 No second ripgrep execution needed!")
        speedup = step1_time / step2_time if step2_time > 0 else float("inf")
        print(f"   📈 Speedup: {speedup:.1f}x faster than re-searching")
    elif summary_result.get("cache_hit"):
        print("   ✅ Direct cache hit (result was already cached)")
    else:
        print("   ⚠️  New search executed (cache derivation not available)")

    print(f"   📁 Files found: {summary_result.get('file_count', 0)}")

    # Show final cache stats
    final_stats = get_default_cache().get_stats()
    print(
        f"   🎯 Final cache stats: {final_stats['hits']} hits, {final_stats['misses']} misses"
    )
    print()

    # Step 3: Demonstrate traditional vs smart caching
    print("📊 Performance Comparison:")
    print(
        f"   Traditional approach: {step1_time:.3f}s + {step1_time:.3f}s = {step1_time * 2:.3f}s"
    )
    print(
        f"   Smart cache approach: {step1_time:.3f}s + {step2_time:.3f}s = {step1_time + step2_time:.3f}s"
    )

    if step2_time < step1_time * 0.1:  # If derivation is much faster
        savings = step1_time - step2_time
        savings_percent = (savings / step1_time) * 100
        print(f"   💰 Time saved: {savings:.3f}s ({savings_percent:.1f}% improvement)")

    print()
    print("🎯 Key Benefits:")
    print("   • Count search provides file list for free")
    print("   • No duplicate ripgrep executions")
    print("   • Optimal for LLM token-efficient workflows")
    print("   • Automatic cross-format result derivation")


if __name__ == "__main__":
    asyncio.run(demo_smart_cache_optimization())



#!/usr/bin/env python3
"""
Base Command Class

Abstract base class for all CLI commands implementing the Command Pattern.
"""

import asyncio
from abc import ABC, abstractmethod
from argparse import Namespace
from typing import Optional

from ...core.analysis_engine import AnalysisRequest, get_analysis_engine
from ...file_handler import read_file_partial
from ...language_detector import detect_language_from_file, is_language_supported
from ...models import AnalysisResult
from ...output_manager import output_error, output_info
from ...project_detector import detect_project_root
from ...security import SecurityValidator


class BaseCommand(ABC):
    """
    Base class for all CLI commands.

    Implements common functionality like file validation, language detection,
    and analysis engine interaction.
    """

    def __init__(self, args: Namespace):
        """Initialize command with parsed arguments."""
        self.args = args

        # Detect project root with priority handling
        file_path = getattr(args, "file_path", None)
        explicit_root = getattr(args, "project_root", None)
        self.project_root = detect_project_root(file_path, explicit_root)

        # Initialize components with project root
        self.analysis_engine = get_analysis_engine(self.project_root)
        self.security_validator = SecurityValidator(self.project_root)

    def validate_file(self) -> bool:
        """Validate input file exists and is accessible."""
        if not hasattr(self.args, "file_path") or not self.args.file_path:
            output_error("File path not specified.")
            return False

        # Security validation
        is_valid, error_msg = self.security_validator.validate_file_path(
            self.args.file_path
        )
        if not is_valid:
            output_error(f"Invalid file path: {error_msg}")
            return False

        from pathlib import Path

        if not Path(self.args.file_path).exists():
            output_error("Invalid file path: file does not exist")
            return False

        return True

    def detect_language(self) -> str | None:
        """Detect or validate the target language."""
        if hasattr(self.args, "language") and self.args.language:
            # Sanitize language input
            sanitized_language = self.security_validator.sanitize_input(
                self.args.language, max_length=50
            )
            target_language = sanitized_language.lower()
            if (not hasattr(self.args, "table") or not self.args.table) and (
                not hasattr(self.args, "quiet") or not self.args.quiet
            ):
                output_info(f"INFO: Language explicitly specified: {target_language}")
        else:
            target_language = detect_language_from_file(self.args.file_path)
            if target_language == "unknown":
                output_error(
                    f"ERROR: Could not determine language for file '{self.args.file_path}'."
                )
                return None
            else:
                if (not hasattr(self.args, "table") or not self.args.table) and (
                    not hasattr(self.args, "quiet") or not self.args.quiet
                ):
                    # Language auto-detected - only show in verbose mode
                    pass

        # Language support validation
        if not is_language_supported(target_language):
            if target_language != "java":
                if (not hasattr(self.args, "table") or not self.args.table) and (
                    not hasattr(self.args, "quiet") or not self.args.quiet
                ):
                    output_info(
                        "INFO: Trying with Java analysis engine. May not work correctly."
                    )
                target_language = "java"  # Fallback

        return str(target_language) if target_language else None

    async def analyze_file(self, language: str) -> Optional["AnalysisResult"]:
        """Perform file analysis using the unified analysis engine."""
        try:
            # Handle partial read if enabled
            if hasattr(self.args, "partial_read") and self.args.partial_read:
                try:
                    partial_content = read_file_partial(
                        self.args.file_path,
                        start_line=self.args.start_line,
                        end_line=getattr(self.args, "end_line", None),
                        start_column=getattr(self.args, "start_column", None),
                        end_column=getattr(self.args, "end_column", None),
                    )
                    if partial_content is None:
                        output_error("Failed to read file partially")
                        return None
                except Exception as e:
                    output_error(f"Failed to read file partially: {e}")
                    return None

            request = AnalysisRequest(
                file_path=self.args.file_path,
                language=language,
                include_complexity=True,
                include_details=True,
            )
            analysis_result = await self.analysis_engine.analyze(request)

            if not analysis_result or not analysis_result.success:
                error_msg = (
                    analysis_result.error_message
                    if analysis_result
                    else "Unknown error"
                )
                output_error(f"Analysis failed: {error_msg}")
                return None

            return analysis_result

        except Exception as e:
            output_error(f"An error occurred during analysis: {e}")
            return None

    def execute(self) -> int:
        """
        Execute the command.

        Returns:
            int: Exit code (0 for success, 1 for failure)
        """
        # Validate inputs
        if not self.validate_file():
            return 1

        # Detect language
        language = self.detect_language()
        if not language:
            return 1

        # Execute the specific command
        try:
            return asyncio.run(self.execute_async(language))
        except Exception as e:
            output_error(f"An error occurred during command execution: {e}")
            return 1

    @abstractmethod
    async def execute_async(self, language: str) -> int:
        """
        Execute the command asynchronously.

        Args:
            language: Detected/specified target language

        Returns:
            int: Exit code (0 for success, 1 for failure)
        """
        pass



#!/usr/bin/env python3
"""
Interactive PyPI Upload Script for tree-sitter-analyzer
Automatically detects version from package metadata
"""

import getpass
import subprocess
import sys
from pathlib import Path

import toml


def get_version() -> str | None:
    """Get version from pyproject.toml"""
    try:
        with open("pyproject.toml", encoding="utf-8") as f:
            data = toml.load(f)
        return data["project"]["version"]
    except Exception as e:
        print(f"❌ Failed to read version from pyproject.toml: {e}")
        return None


def check_packages() -> bool:
    """Check if packages are built"""
    version = get_version()
    if not version:
        return False

    dist_path = Path("dist")
    if not dist_path.exists():
        print("❌ dist/ directory not found. Please run 'uv build' first.")
        return False

    wheel_file = dist_path / f"tree_sitter_analyzer-{version}-py3-none-any.whl"
    tar_file = dist_path / f"tree_sitter_analyzer-{version}.tar.gz"

    if not wheel_file.exists() or not tar_file.exists():
        print(
            f"❌ Package files not found for version {version}. Please run 'uv build' first."
        )
        return False

    print("✅ Package files found:")
    print(f"  - {wheel_file}")
    print(f"  - {tar_file}")
    return True


def upload_with_uv() -> bool:
    """Upload using uv"""
    print("\n🚀 Uploading to PyPI using uv...")

    # Ask for token
    token = getpass.getpass("Enter your PyPI API token (starts with 'pypi-'): ")
    if not token.startswith("pypi-"):
        print("❌ Invalid token format. Token should start with 'pypi-'")
        return False

    try:
        # Upload using uv
        cmd = ["uv", "publish", "--token", token]
        print("Running: uv publish --token [HIDDEN]")

        result = subprocess.run(cmd, capture_output=True, text=True)

        if result.returncode == 0:
            print("✅ Successfully uploaded to PyPI!")
            print(result.stdout)
            return True
        else:
            print("❌ Upload failed:")
            print(result.stderr)
            return False

    except Exception as e:
        print(f"❌ Error during upload: {e}")
        return False


def upload_with_twine() -> bool:
    """Upload using twine"""
    print("\n🚀 Uploading to PyPI using twine...")

    try:
        # Check if twine is available
        subprocess.check_call(
            [sys.executable, "-m", "twine", "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        print("Installing twine...")
        try:
            subprocess.check_call(["uv", "add", "--dev", "twine"])
        except subprocess.CalledProcessError:
            print("❌ Failed to install twine")
            return False

    try:
        # Upload using twine
        cmd = [sys.executable, "-m", "twine", "upload", "dist/*"]
        print("Running: python -m twine upload dist/*")
        print("When prompted:")
        print("  Username: __token__")
        print("  Password: [your PyPI token]")

        result = subprocess.run(cmd)

        if result.returncode == 0:
            print("✅ Successfully uploaded to PyPI!")
            return True
        else:
            print("❌ Upload failed")
            return False

    except Exception as e:
        print(f"❌ Error during upload: {e}")
        return False


def test_installation() -> None:
    """Test installation from PyPI"""
    version = get_version()
    print("\n🧪 Testing installation from PyPI...")
    print("You can test the installation with:")
    print(f"  pip install tree-sitter-analyzer=={version}")
    print(
        '  python -c "import tree_sitter_analyzer; print(tree_sitter_analyzer.__version__)"'
    )


def main() -> None:
    """Main function"""
    version = get_version()
    if not version:
        sys.exit(1)

    print(f"=== Interactive PyPI Upload for tree-sitter-analyzer v{version} ===")
    print()

    # Check packages
    if not check_packages():
        sys.exit(1)

    print("\n📋 Pre-upload checklist:")
    print("✅ All tests passed")
    print("✅ Package built and verified")
    print(f"✅ Version updated to {version}")
    print("✅ Documentation updated")
    print("✅ GitHub tagged and pushed")

    # Add version-specific notes
    if version >= "0.9.3":
        print("✅ CLI output experience improved")

    print("\n🔑 PyPI Account Setup:")
    print("1. Create account: https://pypi.org/account/register/")
    print("2. Generate API token: https://pypi.org/manage/account/token/")
    print("3. Select 'Entire account' scope")
    print("4. Copy the token (starts with 'pypi-')")

    print("\n📦 Upload Options:")
    print("1. Upload with uv (recommended)")
    print("2. Upload with twine")
    print("3. Show manual commands")
    print("4. Exit")

    while True:
        choice = input("\nChoose an option (1-4): ").strip()

        if choice == "1":
            if upload_with_uv():
                test_installation()
                break
        elif choice == "2":
            if upload_with_twine():
                test_installation()
                break
        elif choice == "3":
            version = get_version()
            print("\n📝 Manual Upload Commands:")
            print("Using uv:")
            print("  set UV_PUBLISH_TOKEN=pypi-your-token-here")
            print("  uv publish")
            print()
            print("Using twine:")
            print(f"  uv run twine upload dist/tree_sitter_analyzer-{version}*")
            print("  Username: __token__")
            print("  Password: pypi-your-token-here")
            break
        elif choice == "4":
            print("Exiting...")
            break
        else:
            print("Invalid choice. Please enter 1-4.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Language Plugins

This package contains the core language plugins for the tree-sitter analyzer.
Each plugin provides language-specific parsing and element extraction capabilities.
"""

# This file makes the languages directory a Python package
# Language plugins are automatically discovered by the PluginManager



#!/usr/bin/env python3
"""
Standalone CLI for search_content (ripgrep wrapper)

Maps CLI flags to the MCP SearchContentTool and prints JSON/text.
"""

from __future__ import annotations

import argparse
import asyncio
import sys
from typing import Any

from ...mcp.tools.search_content_tool import SearchContentTool
from ...output_manager import output_data, output_error, set_output_mode
from ...project_detector import detect_project_root


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Search text content in files using ripgrep via MCP wrapper.",
    )

    roots_or_files = parser.add_mutually_exclusive_group(required=True)
    roots_or_files.add_argument(
        "--roots",
        nargs="+",
        help="Directory roots to search recursively",
    )
    roots_or_files.add_argument(
        "--files",
        nargs="+",
        help="Explicit file list to search",
    )

    parser.add_argument("--query", required=True, help="Search pattern")

    # Output
    parser.add_argument(
        "--output-format",
        choices=["json", "text", "toon"],
        default="json",
        help="Output format: 'json' (default), 'text', or 'toon' (50-70%% token reduction)",
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Suppress non-essential output",
    )

    # rg options
    parser.add_argument(
        "--case", choices=["smart", "insensitive", "sensitive"], default="smart"
    )
    parser.add_argument("--fixed-strings", action="store_true")
    parser.add_argument("--word", action="store_true")
    parser.add_argument("--multiline", action="store_true")
    parser.add_argument("--include-globs", nargs="+")
    parser.add_argument("--exclude-globs", nargs="+")
    parser.add_argument("--follow-symlinks", action="store_true")
    parser.add_argument("--hidden", action="store_true")
    parser.add_argument("--no-ignore", action="store_true")
    parser.add_argument("--max-filesize")
    parser.add_argument("--context-before", type=int)
    parser.add_argument("--context-after", type=int)
    parser.add_argument("--encoding")
    parser.add_argument("--max-count", type=int)
    parser.add_argument("--timeout-ms", type=int)
    parser.add_argument("--count-only-matches", action="store_true")
    parser.add_argument("--summary-only", action="store_true")
    parser.add_argument("--optimize-paths", action="store_true")
    parser.add_argument("--group-by-file", action="store_true")
    parser.add_argument("--total-only", action="store_true")

    # project root
    parser.add_argument(
        "--project-root",
        help="Project root directory for security boundary (auto-detected if omitted)",
    )

    return parser


async def _run(args: argparse.Namespace) -> int:
    set_output_mode(quiet=bool(args.quiet), json_output=(args.output_format == "json"))

    project_root = detect_project_root(None, args.project_root)
    tool = SearchContentTool(project_root)

    payload: dict[str, Any] = {
        "query": args.query,
    }
    if args.roots:
        payload["roots"] = list(args.roots)
    if args.files:
        payload["files"] = list(args.files)

    # Options mapping
    if args.case:
        payload["case"] = args.case
    if args.fixed_strings:
        payload["fixed_strings"] = True
    if args.word:
        payload["word"] = True
    if args.multiline:
        payload["multiline"] = True
    if args.include_globs:
        payload["include_globs"] = args.include_globs
    if args.exclude_globs:
        payload["exclude_globs"] = args.exclude_globs
    if args.follow_symlinks:
        payload["follow_symlinks"] = True
    if args.hidden:
        payload["hidden"] = True
    if args.no_ignore:
        payload["no_ignore"] = True
    if args.max_filesize:
        payload["max_filesize"] = args.max_filesize
    if args.context_before is not None:
        payload["context_before"] = int(args.context_before)
    if args.context_after is not None:
        payload["context_after"] = int(args.context_after)
    if args.encoding:
        payload["encoding"] = args.encoding
    if args.max_count is not None:
        payload["max_count"] = int(args.max_count)
    if args.timeout_ms is not None:
        payload["timeout_ms"] = int(args.timeout_ms)
    if args.count_only_matches:
        payload["count_only_matches"] = True
    if args.summary_only:
        payload["summary_only"] = True
    if args.optimize_paths:
        payload["optimize_paths"] = True
    if args.group_by_file:
        payload["group_by_file"] = True
    if args.total_only:
        payload["total_only"] = True

    # Pass output_format to MCP tool to ensure consistent formatting
    # CLI uses "json" as default, MCP tool uses "toon" as default
    # We need to explicitly pass the CLI's output_format to override MCP default
    payload["output_format"] = args.output_format

    try:
        result = await tool.execute(payload)
        output_data(result, args.output_format)
        return 0 if (isinstance(result, dict) or isinstance(result, int)) else 0
    except Exception as e:
        output_error(str(e))
        return 1


def main() -> None:
    parser = _build_parser()
    args = parser.parse_args()
    try:
        rc = asyncio.run(_run(args))
    except KeyboardInterrupt:
        rc = 1
    sys.exit(rc)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
TOON (Token-Oriented Object Notation) Formatter

High-level formatter for converting analysis results and MCP responses
to TOON format, optimized for LLM consumption with 50-70% token reduction.
"""

import logging
from typing import Any

from .base_formatter import BaseFormatter
from .toon_encoder import ToonEncodeError, ToonEncoder

# Logger for TOON formatter
logger = logging.getLogger(__name__)


class ToonFormatter(BaseFormatter):
    """
    TOON formatter for LLM-optimized output.

    Converts analysis results to compact, human-readable TOON format that
    reduces token consumption by 50-70% compared to JSON while maintaining
    full information fidelity.

    Implements the unified Formatter interface for OutputManager compatibility.
    """

    def __init__(
        self,
        use_tabs: bool = False,
        compact_arrays: bool = True,
        include_metadata: bool = True,
        fallback_to_json: bool = True,
        normalize_paths: bool = True,
    ):
        """
        Initialize TOON formatter.

        Args:
            use_tabs: Use tab delimiters instead of commas (further optimization)
            compact_arrays: Use CSV-style compact arrays for homogeneous data
            include_metadata: Include file metadata in output
            fallback_to_json: Fall back to JSON on encoding errors
            normalize_paths: Convert Windows backslashes to forward slashes
                           for ~10% token reduction in path-heavy outputs
        """
        self.use_tabs = use_tabs
        self.compact_arrays = compact_arrays
        self.include_metadata = include_metadata
        self.fallback_to_json = fallback_to_json
        self.normalize_paths = normalize_paths
        self.encoder = ToonEncoder(
            use_tabs=use_tabs,
            fallback_to_json=fallback_to_json,
            normalize_paths=normalize_paths,
        )

    def format(self, data: Any) -> str:
        """
        Unified format method implementing the Formatter protocol.

        Routes to appropriate internal formatter based on data type.
        This method enables OutputManager to call formatter.format(data)
        without needing to know the specific formatter implementation.

        On encoding errors, falls back to JSON if fallback_to_json is True.

        Args:
            data: The data to format (AnalysisResult, dict, or other types)

        Returns:
            TOON-formatted string (or JSON on fallback)
        """
        try:
            return self._format_internal(data)
        except ToonEncodeError as e:
            logger.error(f"TOON formatting failed: {e}")
            if self.fallback_to_json:
                logger.warning("Falling back to JSON format")
                return self.encoder._fallback_to_json(data)
            raise
        except Exception as e:
            logger.error(f"Unexpected error during TOON formatting: {e}", exc_info=True)
            if self.fallback_to_json:
                logger.warning("Falling back to JSON format")
                return self.encoder._fallback_to_json(data)
            raise ToonEncodeError("Formatting failed", data=data, cause=e) from e

    def _format_internal(self, data: Any) -> str:
        """
        Internal format method without error handling wrapper.

        Args:
            data: The data to format

        Returns:
            TOON-formatted string
        """
        # Import here to avoid circular dependency
        try:
            from tree_sitter_analyzer.models import AnalysisResult

            if isinstance(data, AnalysisResult):
                return self.format_analysis_result(data)
        except ImportError:
            pass

        # Check if it's an MCP response (dict with specific structure)
        if isinstance(data, dict):
            # Detect MCP response structure
            if self._is_mcp_response(data):
                return self.format_mcp_response(data)
            else:
                # Generic dict - try to format as analysis result dict
                return self.format_structure(data)

        # Fallback: encode arbitrary data as TOON
        return self.encoder.encode(data)

    def _is_mcp_response(self, data: dict[str, Any]) -> bool:
        """
        Detect if data is an MCP response structure.

        MCP responses typically have 'content' or 'data' fields.

        Args:
            data: Dictionary to check

        Returns:
            True if data appears to be an MCP response
        """
        mcp_keys = {"content", "data", "metadata", "analysis_result"}
        return bool(mcp_keys.intersection(data.keys()))

    def format_analysis_result(self, result: Any, table_type: str = "full") -> str:
        """
        Format complete analysis result as TOON.

        Args:
            result: Analysis result to format (AnalysisResult object)
            table_type: Output detail level (full, summary, compact)

        Returns:
            TOON-formatted string
        """
        # Import here to avoid circular dependency
        from tree_sitter_analyzer.models import AnalysisResult

        if not isinstance(result, AnalysisResult):
            # Try to convert dict to proper format
            if isinstance(result, dict):
                return self.format_structure(result)
            return self.encoder.encode(result)

        lines = []

        # File metadata
        if self.include_metadata:
            lines.append(f"file: {result.file_path}")
            lines.append(f"language: {result.language}")
            if result.package:
                package_name = (
                    result.package.name
                    if hasattr(result.package, "name")
                    else str(result.package)
                )
                lines.append(f"package: {package_name}")
            lines.append("")

        # Statistics summary
        stats = result.get_summary()
        if stats:
            lines.append("summary:")
            for key, value in stats.items():
                lines.append(f"  {key}: {value}")
            lines.append("")

        # Elements (classes, methods, functions)
        if result.elements:
            lines.append(f"elements[{len(result.elements)}]:")

            # Group by type
            classes = [e for e in result.elements if e.element_type == "class"]
            methods = [e for e in result.elements if e.element_type == "method"]
            functions = [e for e in result.elements if e.element_type == "function"]

            if classes:
                lines.append(f"  classes[{len(classes)}]:")
                for cls in classes:
                    lines.append(f"    - {cls.name}")

            if methods:
                lines.append(f"  methods[{len(methods)}]:")
                if self.compact_arrays:
                    # Use compact table format
                    method_dicts = [
                        self._method_to_dict(m) for m in methods[:10]
                    ]  # Limit for demo
                    table = self.encoder.encode_array_table(method_dicts, indent=2)
                    lines.append(table)
                else:
                    for method in methods[:10]:  # Limit for demo
                        lines.append(f"    - {method.name}")

            if functions:
                lines.append(f"  functions[{len(functions)}]:")
                for func in functions[:10]:  # Limit for demo
                    lines.append(f"    - {func.name}")

        return "\n".join(lines)

    def format_mcp_response(self, data: dict[str, Any]) -> str:
        """
        Format MCP tool response as TOON.

        Optimized for AI assistant consumption.

        Args:
            data: MCP response dictionary

        Returns:
            TOON-formatted string
        """
        return self.encoder.encode_dict(data)

    def format_summary(self, analysis_result: dict[str, Any]) -> str:
        """
        Format summary output (BaseFormatter requirement).

        Args:
            analysis_result: Analysis result dictionary

        Returns:
            TOON-formatted summary
        """
        return self.encoder.encode_dict(analysis_result)

    def format_structure(self, analysis_result: dict[str, Any]) -> str:
        """
        Format structure analysis output (BaseFormatter requirement).

        Args:
            analysis_result: Analysis result dictionary

        Returns:
            TOON-formatted structure
        """
        return self.encoder.encode_dict(analysis_result)

    def format_advanced(
        self, analysis_result: dict[str, Any], output_format: str = "toon"
    ) -> str:
        """
        Format advanced analysis output (BaseFormatter requirement).

        Args:
            analysis_result: Analysis result dictionary
            output_format: Output format (ignored, always returns TOON)

        Returns:
            TOON-formatted advanced analysis
        """
        return self.encoder.encode_dict(analysis_result)

    def format_table(
        self, analysis_result: dict[str, Any], table_type: str = "full"
    ) -> str:
        """
        Format table output (BaseFormatter requirement).

        Args:
            analysis_result: Analysis result dictionary
            table_type: Table detail level (full, compact, summary)

        Returns:
            TOON-formatted table
        """
        return self.encoder.encode_dict(analysis_result)

    def _method_to_dict(self, method: Any) -> dict[str, Any]:
        """
        Convert method element to dictionary for table encoding.

        Args:
            method: Method element

        Returns:
            Dictionary representation
        """
        return {
            "name": method.name if hasattr(method, "name") else str(method),
            "visibility": getattr(method, "visibility", ""),
            "lines": f"{getattr(method, 'start_line', 0)}-{getattr(method, 'end_line', 0)}",
        }

    @staticmethod
    def is_toon_content(content: str) -> bool:
        """
        Detect if content is in TOON format.

        Used by FileOutputManager to determine content type.

        Args:
            content: String content to check

        Returns:
            True if content appears to be TOON format
        """
        if not content or not content.strip():
            return False

        lines = content.strip().split("\n")
        if not lines:
            return False

        # TOON format indicators:
        # 1. Lines with "key: value" pattern (not JSON)
        # 2. Array table headers like "[N]{field1,field2}:"
        # 3. Nested structure with indentation

        # Check if it looks like JSON first
        first_char = content.strip()[0]
        if first_char in "{[":
            return False

        # Check for TOON patterns
        toon_patterns = 0
        for line in lines[:10]:  # Check first 10 lines
            stripped = line.strip()
            if not stripped:
                continue

            # Key-value pattern: "key: value"
            if (
                ":" in stripped
                and not stripped.startswith("{")
                and not stripped.startswith('"')
            ):
                parts = stripped.split(":", 1)
                if len(parts) == 2 and parts[0].strip():
                    toon_patterns += 1

            # Array table header: "[N]{...}:"
            if stripped.startswith("[") and "]{" in stripped and stripped.endswith(":"):
                toon_patterns += 2

        # Need at least 2 TOON patterns to confirm
        return toon_patterns >= 2



#!/usr/bin/env python3
"""
Markdown Query Definitions

Tree-sitter queries for extracting Markdown elements including headers,
links, code blocks, lists, and other structural elements.
"""

# Markdown element extraction queries - simplified for compatibility
MARKDOWN_QUERIES: dict[str, str] = {
    # Headers (H1-H6) - simplified
    "headers": """
    (atx_heading) @header
    (setext_heading) @header
    """,
    # Code blocks - simplified
    "code_blocks": """
    (fenced_code_block) @code_block
    (indented_code_block) @code_block
    """,
    # Inline code - simplified
    "inline_code": """
    (inline) @inline
    """,
    # Links - simplified to avoid invalid node types
    "links": """
    (inline) @inline
    """,
    # Images - simplified to avoid invalid node types
    "images": """
    (inline) @inline
    """,
    # Lists - simplified to avoid invalid node types
    "lists": """
    (list) @list
    (list_item) @list_item
    """,
    # Emphasis and strong - simplified
    "emphasis": """
    (inline) @inline
    """,
    # Blockquotes - simplified
    "blockquotes": """
    (block_quote) @blockquote
    """,
    # Tables - simplified
    "tables": """
    (pipe_table) @table
    """,
    # Horizontal rules - simplified
    "horizontal_rules": """
    (thematic_break) @hr
    """,
    # HTML blocks - simplified
    "html_blocks": """
    (html_block) @html_block
    """,
    # Inline HTML - simplified
    "inline_html": """
    (inline) @inline
    """,
    # Strikethrough - simplified
    "strikethrough": """
    (inline) @inline
    """,
    # Task lists - simplified
    "task_lists": """
    (list_item) @list_item
    """,
    # Footnotes - simplified
    "footnotes": """
    (paragraph) @paragraph
    (inline) @inline
    """,
    # All text content - simplified
    "text_content": """
    (paragraph) @paragraph
    (inline) @inline
    """,
    # Document structure - simplified
    "document": """
    (document) @document
    """,
    # All elements (comprehensive) - simplified
    "all_elements": """
    (atx_heading) @heading
    (setext_heading) @heading
    (fenced_code_block) @code_block
    (indented_code_block) @code_block
    (inline) @inline
    (list) @list
    (list_item) @list_item
    (block_quote) @blockquote
    (pipe_table) @table
    (thematic_break) @hr
    (html_block) @html_block
    (paragraph) @paragraph
    (link_reference_definition) @reference
    """,
}

# Query aliases for convenience
QUERY_ALIASES: dict[str, str] = {
    "heading": "headers",
    "h1": "headers",
    "h2": "headers",
    "h3": "headers",
    "h4": "headers",
    "h5": "headers",
    "h6": "headers",
    "code": "code_blocks",
    "fenced_code": "code_blocks",
    "code_span": "inline_code",
    "link": "links",
    "url": "links",
    "image": "images",
    "img": "images",
    "list": "lists",
    "ul": "lists",
    "ol": "lists",
    "em": "emphasis",
    "strong": "emphasis",
    "bold": "emphasis",
    "italic": "emphasis",
    "quote": "blockquotes",
    "blockquote": "blockquotes",
    "table": "tables",
    "hr": "horizontal_rules",
    "html": "html_blocks",
    "strike": "strikethrough",
    "task": "task_lists",
    "todo": "task_lists",
    "footnote": "footnotes",
    "note": "footnotes",
    "text": "text_content",
    "paragraph": "text_content",
    "all": "all_elements",
    "everything": "all_elements",
}


def get_query(query_name: str) -> str:
    """
    Get a query by name, supporting aliases

    Args:
        query_name: Name of the query or alias

    Returns:
        Query string

    Raises:
        KeyError: If query name is not found
    """
    # Check direct queries first
    if query_name in MARKDOWN_QUERIES:
        return MARKDOWN_QUERIES[query_name]

    # Check aliases
    if query_name in QUERY_ALIASES:
        actual_query = QUERY_ALIASES[query_name]
        return MARKDOWN_QUERIES[actual_query]

    raise KeyError(f"Unknown query: {query_name}")


def get_available_queries() -> list[str]:
    """
    Get list of all available query names including aliases

    Returns:
        List of query names
    """
    queries = list(MARKDOWN_QUERIES.keys())
    aliases = list(QUERY_ALIASES.keys())
    return sorted(queries + aliases)


def get_query_info(query_name: str) -> dict[str, str | bool]:
    """
    Get information about a query

    Args:
        query_name: Name of the query

    Returns:
        Dictionary with query information
    """
    try:
        query_string = get_query(query_name)
        is_alias = query_name in QUERY_ALIASES
        actual_name = (
            QUERY_ALIASES.get(query_name, query_name) if is_alias else query_name
        )

        return {
            "name": query_name,
            "actual_name": actual_name,
            "is_alias": is_alias,
            "query": query_string,
            "description": _get_query_description(actual_name),
        }
    except KeyError:
        return {"error": f"Query '{query_name}' not found"}


def _get_query_description(query_name: str) -> str:
    """Get description for a query"""
    descriptions = {
        "headers": "Extract all heading elements (H1-H6, both ATX and Setext styles)",
        "code_blocks": "Extract fenced and indented code blocks",
        "inline_code": "Extract inline code spans",
        "links": "Extract all types of links (inline, reference, autolinks)",
        "images": "Extract image elements (inline and reference)",
        "lists": "Extract ordered and unordered lists",
        "emphasis": "Extract emphasis and strong emphasis elements",
        "blockquotes": "Extract blockquote elements",
        "tables": "Extract pipe table elements",
        "horizontal_rules": "Extract horizontal rule elements",
        "html_blocks": "Extract HTML block elements",
        "inline_html": "Extract inline HTML elements",
        "strikethrough": "Extract strikethrough elements",
        "task_lists": "Extract task list items (checkboxes)",
        "footnotes": "Extract footnote references and definitions",
        "text_content": "Extract all text content",
        "document": "Extract document root",
        "all_elements": "Extract all Markdown elements",
    }
    return descriptions.get(query_name, "No description available")


def get_all_queries() -> dict[str, str]:
    """
    Get all queries for the query loader

    Returns:
        Dictionary mapping query names to query strings
    """
    # Combine direct queries and aliases
    all_queries = MARKDOWN_QUERIES.copy()

    # Add aliases that point to actual queries
    for alias, target in QUERY_ALIASES.items():
        if target in MARKDOWN_QUERIES:
            all_queries[alias] = MARKDOWN_QUERIES[target]

    return all_queries


# Export main functions and constants
__all__ = [
    "MARKDOWN_QUERIES",
    "QUERY_ALIASES",
    "get_query",
    "get_available_queries",
    "get_query_info",
    "get_all_queries",
]



#!/usr/bin/env python3
"""
C++ Language Queries

Tree-sitter queries specific to C++ language constructs.
Covers classes, functions, templates, namespaces, and other C++ specific elements.
"""

# C++-specific query library
CPP_QUERIES: dict[str, str] = {
    # --- Basic Structure ---
    "class": """
    (class_specifier) @class
    """,
    "struct": """
    (struct_specifier) @struct
    """,
    "union": """
    (union_specifier) @union
    """,
    "enum": """
    (enum_specifier) @enum
    """,
    # --- Functions and Methods ---
    "function": """
    (function_definition) @function
    """,
    "function_declaration": """
    (declaration
      declarator: (function_declarator)) @function_declaration
    """,
    "method": """
    (function_definition
      declarator: (function_declarator
        declarator: (field_identifier) @method.name)) @method
    """,
    "constructor": """
    (function_definition
      declarator: (function_declarator
        declarator: (identifier) @ctor.name)
      body: (compound_statement)) @constructor
    """,
    "destructor": """
    (function_definition
      declarator: (function_declarator
        declarator: (destructor_name) @dtor.name)) @destructor
    """,
    "virtual_function": """
    (function_definition
      (virtual)) @virtual_function
    """,
    # --- Templates ---
    "template": """
    (template_declaration) @template
    """,
    "template_function": """
    (template_declaration
      (function_definition) @template_function)
    """,
    "template_class": """
    (template_declaration
      (class_specifier) @template_class)
    """,
    "template_parameter": """
    (template_parameter_list
      (type_parameter_declaration) @template_parameter)
    """,
    # --- Namespaces ---
    "namespace": """
    (namespace_definition) @namespace
    """,
    "namespace_name": """
    (namespace_definition
      name: (identifier) @namespace_name)
    """,
    "using_declaration": """
    (using_declaration) @using_declaration
    """,
    "using_directive": """
    (using_directive) @using_directive
    """,
    # --- Includes and Preprocessor ---
    "include": """
    (preproc_include) @include
    """,
    "system_include": """
    (preproc_include
      path: (system_lib_string) @system_include)
    """,
    "local_include": """
    (preproc_include
      path: (string_literal) @local_include)
    """,
    "define": """
    (preproc_def) @define
    """,
    "ifdef": """
    (preproc_ifdef) @ifdef
    """,
    "macro_function": """
    (preproc_function_def) @macro_function
    """,
    # --- Variables and Fields ---
    "field": """
    (field_declaration) @field
    """,
    "variable": """
    (declaration) @variable
    """,
    "static_variable": """
    (declaration
      (storage_class_specifier) @storage
      (#eq? @storage "static")) @static_variable
    """,
    "const_variable": """
    (declaration
      (type_qualifier) @qualifier
      (#eq? @qualifier "const")) @const_variable
    """,
    # --- Access Specifiers ---
    "public_access": """
    (access_specifier) @public_access
    (#eq? @public_access "public:")
    """,
    "private_access": """
    (access_specifier) @private_access
    (#eq? @private_access "private:")
    """,
    "protected_access": """
    (access_specifier) @protected_access
    (#eq? @protected_access "protected:")
    """,
    # --- Inheritance ---
    "base_class": """
    (base_class_clause
      (base_specifier) @base_class)
    """,
    "public_inheritance": """
    (base_class_clause
      (base_specifier
        (access_specifier) @access
        (#eq? @access "public")
        (type_identifier) @base_type)) @public_inheritance
    """,
    # --- Smart Pointers and Modern C++ ---
    "smart_pointer": """
    (template_type
      name: (type_identifier) @ptr_type
      (#match? @ptr_type "^(unique_ptr|shared_ptr|weak_ptr)$")) @smart_pointer
    """,
    "auto_type": """
    (declaration
      type: (auto)) @auto_type
    """,
    "lambda": """
    (lambda_expression) @lambda
    """,
    "range_for": """
    (for_range_loop) @range_for
    """,
    # --- Exception Handling ---
    "try_catch": """
    (try_statement) @try_catch
    """,
    "catch_clause": """
    (catch_clause) @catch_clause
    """,
    "throw_statement": """
    (throw_statement) @throw_statement
    """,
    # --- Name-only Extraction ---
    "class_name": """
    (class_specifier
      name: (type_identifier) @class_name)
    """,
    "function_name": """
    (function_definition
      declarator: (function_declarator
        declarator: (identifier) @function_name))
    """,
    "field_name": """
    (field_declaration
      declarator: (field_identifier) @field_name)
    """,
    # --- Comments ---
    "comment": """
    (comment) @comment
    """,
    "block_comment": """
    (comment) @block_comment
    (#match? @block_comment "^/\\*")
    """,
    # --- Operator Overloading ---
    "operator_overload": """
    (function_definition
      declarator: (function_declarator
        declarator: (operator_name) @operator_name)) @operator_overload
    """,
    # --- Friend Declaration ---
    "friend": """
    (friend_declaration) @friend
    """,
}

# Query descriptions
CPP_QUERY_DESCRIPTIONS: dict[str, str] = {
    "class": "Extract C++ class declarations",
    "struct": "Extract C++ struct declarations",
    "union": "Extract C++ union declarations",
    "enum": "Extract C++ enum declarations",
    "function": "Extract C++ function definitions",
    "function_declaration": "Extract C++ function declarations",
    "method": "Extract C++ class methods",
    "constructor": "Extract C++ constructors",
    "destructor": "Extract C++ destructors",
    "virtual_function": "Extract C++ virtual functions",
    "template": "Extract C++ template declarations",
    "template_function": "Extract C++ template functions",
    "template_class": "Extract C++ template classes",
    "template_parameter": "Extract C++ template parameters",
    "namespace": "Extract C++ namespace definitions",
    "namespace_name": "Extract C++ namespace names",
    "using_declaration": "Extract C++ using declarations",
    "using_directive": "Extract C++ using directives",
    "include": "Extract C++ include directives",
    "system_include": "Extract C++ system includes",
    "local_include": "Extract C++ local includes",
    "define": "Extract C++ preprocessor defines",
    "ifdef": "Extract C++ preprocessor ifdef blocks",
    "macro_function": "Extract C++ macro functions",
    "field": "Extract C++ class/struct fields",
    "variable": "Extract C++ variable declarations",
    "static_variable": "Extract C++ static variables",
    "const_variable": "Extract C++ const variables",
    "public_access": "Extract C++ public access specifiers",
    "private_access": "Extract C++ private access specifiers",
    "protected_access": "Extract C++ protected access specifiers",
    "base_class": "Extract C++ base class specifiers",
    "public_inheritance": "Extract C++ public inheritance",
    "smart_pointer": "Extract C++ smart pointer usage",
    "auto_type": "Extract C++ auto type declarations",
    "lambda": "Extract C++ lambda expressions",
    "range_for": "Extract C++ range-based for loops",
    "try_catch": "Extract C++ try-catch blocks",
    "catch_clause": "Extract C++ catch clauses",
    "throw_statement": "Extract C++ throw statements",
    "class_name": "Extract C++ class names only",
    "function_name": "Extract C++ function names only",
    "field_name": "Extract C++ field names only",
    "comment": "Extract C++ comments",
    "block_comment": "Extract C++ block comments",
    "operator_overload": "Extract C++ operator overloads",
    "friend": "Extract C++ friend declarations",
}


def get_cpp_query(name: str) -> str:
    """
    Get the specified C++ query

    Args:
        name: Query name

    Returns:
        Query string

    Raises:
        ValueError: When query is not found
    """
    if name not in CPP_QUERIES:
        available = list(CPP_QUERIES.keys())
        raise ValueError(f"C++ query '{name}' does not exist. Available: {available}")

    return CPP_QUERIES[name]


def get_cpp_query_description(name: str) -> str:
    """
    Get the description of the specified C++ query

    Args:
        name: Query name

    Returns:
        Query description
    """
    return CPP_QUERY_DESCRIPTIONS.get(name, "No description")


# Convert to ALL_QUERIES format for dynamic loader compatibility
ALL_QUERIES = {}
for query_name, query_string in CPP_QUERIES.items():
    description = CPP_QUERY_DESCRIPTIONS.get(query_name, "No description")
    ALL_QUERIES[query_name] = {"query": query_string, "description": description}

# Add common query aliases for cross-language compatibility
ALL_QUERIES["functions"] = {
    "query": CPP_QUERIES["function"],
    "description": "Search all function definitions (alias for function)",
}

ALL_QUERIES["methods"] = {
    "query": CPP_QUERIES["method"],
    "description": "Search all method declarations (alias for method)",
}

ALL_QUERIES["classes"] = {
    "query": CPP_QUERIES["class"],
    "description": "Search all class declarations (alias for class)",
}

ALL_QUERIES["imports"] = {
    "query": CPP_QUERIES["include"],
    "description": "Search all include directives (alias for include)",
}

ALL_QUERIES["variables"] = {
    "query": CPP_QUERIES["variable"],
    "description": "Search all variable declarations (alias for variable)",
}


def get_query(name: str) -> str:
    """Get a specific query by name."""
    if name in ALL_QUERIES:
        return ALL_QUERIES[name]["query"]
    raise ValueError(
        f"Query '{name}' not found. Available queries: {list(ALL_QUERIES.keys())}"
    )


def get_all_queries() -> dict:
    """Get all available queries."""
    return ALL_QUERIES


def list_queries() -> list:
    """List all available query names."""
    return list(ALL_QUERIES.keys())


def get_available_cpp_queries() -> list[str]:
    """
    Get list of available C++ queries

    Returns:
        List of query names
    """
    return list(CPP_QUERIES.keys())



#!/usr/bin/env python3
"""
C# Language Plugin

Provides C#-specific parsing and element extraction functionality.
Supports extraction of classes, interfaces, records, methods, properties, fields, and using directives.
"""

from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    import tree_sitter

    from ..core.analysis_engine import AnalysisRequest
    from ..models import AnalysisResult

try:
    import tree_sitter

    TREE_SITTER_AVAILABLE = True
except ImportError:
    TREE_SITTER_AVAILABLE = False

from ..models import Class, Function, Import, Variable
from ..plugins.base import ElementExtractor, LanguagePlugin
from ..utils import log_debug, log_error


class CSharpElementExtractor(ElementExtractor):
    """
    C#-specific element extractor.

    This extractor parses C# AST and extracts code elements, mapping them
    to the unified element model:
    - Classes, Interfaces, Records, Enums, Structs → Class elements
    - Methods, Constructors, Properties → Function elements
    - Fields, Constants, Events → Variable elements
    - Using directives → Import elements

    The extractor handles modern C# syntax including:
    - C# 8+ nullable reference types
    - C# 9+ records
    - Async/await patterns
    - Attributes (annotations)
    - Generic types
    """

    def __init__(self) -> None:
        """
        Initialize the C# element extractor.

        Sets up internal state for source code processing and performance
        optimization caches for node text extraction.
        """
        super().__init__()
        self.source_code: str = ""
        self.content_lines: list[str] = []
        self.current_namespace: str = ""

        # Performance optimization caches - use position-based keys for deterministic caching
        self._node_text_cache: dict[tuple[int, int], str] = {}
        self._processed_nodes: set[tuple[int, int]] = set()
        self._element_cache: dict[tuple[tuple[int, int], str], Any] = {}
        self._file_encoding: str | None = None
        self._attribute_cache: dict[tuple[int, int], list[dict[str, Any]]] = {}

    def _reset_caches(self) -> None:
        """Reset all internal caches for a new file analysis."""
        self._node_text_cache.clear()
        self._processed_nodes.clear()
        self._element_cache.clear()
        self._attribute_cache.clear()
        self.current_namespace = ""

    def _get_node_text_optimized(self, node: "tree_sitter.Node") -> str:
        """
        Get text content of a node with caching for performance.

        Args:
            node: Tree-sitter node to extract text from

        Returns:
            Text content of the node as string
        """
        # Use node position as cache key instead of object id for deterministic behavior
        cache_key = (node.start_byte, node.end_byte)
        if cache_key in self._node_text_cache:
            return self._node_text_cache[cache_key]

        # Extract text directly from source code string
        text = self.source_code[node.start_byte : node.end_byte]
        self._node_text_cache[cache_key] = text
        return text

    def _extract_namespace(self, node: "tree_sitter.Node") -> None:
        """
        Extract namespace from the AST and set current_namespace.

        Args:
            node: Root node of the AST
        """
        if node.type == "namespace_declaration":
            name_node = node.child_by_field_name("name")
            if name_node:
                self.current_namespace = self._get_node_text_optimized(name_node)
                return

        # Recursively search for namespace
        for child in node.children:
            if child.type == "namespace_declaration":
                name_node = child.child_by_field_name("name")
                if name_node:
                    self.current_namespace = self._get_node_text_optimized(name_node)
                    return
            elif child.child_count > 0:
                self._extract_namespace(child)

    def _extract_modifiers(self, node: "tree_sitter.Node") -> list[str]:
        """
        Extract modifiers from a declaration node.

        Args:
            node: Declaration node (class, method, field, etc.)

        Returns:
            List of modifier strings (e.g., ["public", "static", "async"])
        """
        modifiers: list[str] = []
        for child in node.children:
            if child.type == "modifier":
                modifier_text = self._get_node_text_optimized(child)
                modifiers.append(modifier_text)
        return modifiers

    def _determine_visibility(self, modifiers: list[str]) -> str:
        """
        Determine visibility from modifiers.

        Args:
            modifiers: List of modifier strings

        Returns:
            Visibility string ("public", "private", "protected", "internal")
        """
        if "public" in modifiers:
            return "public"
        elif "private" in modifiers:
            return "private"
        elif "protected" in modifiers:
            return "protected"
        elif "internal" in modifiers:
            return "internal"
        else:
            return "private"  # Default to private in C#

    def _extract_attributes(self, node: "tree_sitter.Node") -> list[dict[str, Any]]:
        """
        Extract attributes (annotations) from a node.

        Args:
            node: Node that may have attributes

        Returns:
            List of attribute dictionaries with name, line, and text
        """
        # Use position-based cache key for deterministic behavior
        cache_key = (node.start_byte, node.end_byte)
        if cache_key in self._attribute_cache:
            return self._attribute_cache[cache_key]

        attributes: list[dict[str, Any]] = []

        # Look for attribute_list nodes before the declaration
        prev_sibling = node.prev_sibling
        while prev_sibling:
            if prev_sibling.type == "attribute_list":
                attr_text = self._get_node_text_optimized(prev_sibling)
                attributes.append(
                    {
                        "name": attr_text.strip("[]"),
                        "line": prev_sibling.start_point[0] + 1,
                        "text": attr_text,
                    }
                )
            elif prev_sibling.type not in ["comment", "line_comment", "block_comment"]:
                break
            prev_sibling = prev_sibling.prev_sibling

        attributes.reverse()  # Restore original order
        self._attribute_cache[cache_key] = attributes
        return attributes

    def _extract_type_name(self, type_node: Optional["tree_sitter.Node"]) -> str:
        """
        Extract type name from a type node, handling generic types and nullable types.

        Args:
            type_node: Type node from the AST

        Returns:
            Type name as string (e.g., "int", "List<string>", "string?")
        """
        if not type_node:
            return "void"

        return self._get_node_text_optimized(type_node)

    def _extract_parameters(
        self, params_node: Optional["tree_sitter.Node"]
    ) -> list[str]:
        """
        Extract method parameters.

        Args:
            params_node: Parameter list node

        Returns:
            List of parameter strings (e.g., ["int id", "string name"])
        """
        if not params_node:
            return []

        parameters: list[str] = []
        for child in params_node.children:
            if child.type == "parameter":
                param_text = self._get_node_text_optimized(child)
                parameters.append(param_text)

        return parameters

    def _traverse_iterative(
        self, root_node: "tree_sitter.Node"
    ) -> Iterator["tree_sitter.Node"]:
        """
        Iteratively traverse AST nodes to avoid stack overflow on large files.

        Args:
            root_node: Root node to start traversal from

        Yields:
            Tree-sitter nodes in depth-first order
        """
        stack = [root_node]
        while stack:
            node = stack.pop()
            yield node
            # Add children in reverse order to maintain left-to-right traversal
            stack.extend(reversed(list(node.children)))

    def extract_classes(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Class]:
        """
        Extract classes, interfaces, records, enums, and structs.

        Args:
            tree: Tree-sitter AST tree parsed from C# source
            source_code: Original C# source code as string

        Returns:
            List of Class objects representing all class-like declarations
        """
        self.source_code = source_code or ""
        self.content_lines = self.source_code.split("\n")
        self._reset_caches()

        classes: list[Class] = []

        if tree is None or tree.root_node is None:
            return classes

        # Extract namespace first
        self._extract_namespace(tree.root_node)

        # Extract all class-like declarations
        for node in self._traverse_iterative(tree.root_node):
            if node.type in [
                "class_declaration",
                "interface_declaration",
                "record_declaration",
                "enum_declaration",
                "struct_declaration",
            ]:
                class_obj = self._extract_class_declaration(node)
                if class_obj:
                    classes.append(class_obj)

        # Sort by start line for deterministic output
        classes.sort(key=lambda c: c.start_line)

        return classes

    def _extract_class_declaration(self, node: "tree_sitter.Node") -> Class | None:
        """
        Extract a single class declaration.

        Args:
            node: Class declaration node

        Returns:
            Class object or None if extraction fails
        """
        try:
            # Get class name
            name_node = node.child_by_field_name("name")
            if not name_node:
                return None

            class_name = self._get_node_text_optimized(name_node)

            # Get modifiers and visibility
            modifiers = self._extract_modifiers(node)
            visibility = self._determine_visibility(modifiers)

            # Get attributes
            attributes = self._extract_attributes(node)

            # Get base class and interfaces
            base_list_node = node.child_by_field_name("bases")
            superclass = None
            interfaces: list[str] = []

            if base_list_node:
                base_items = [
                    self._get_node_text_optimized(child)
                    for child in base_list_node.children
                    if child.type
                    in ["type_identifier", "generic_name", "qualified_name"]
                ]
                if base_items:
                    if node.type == "interface_declaration":
                        interfaces = base_items
                    else:
                        superclass = base_items[0]
                        interfaces = base_items[1:] if len(base_items) > 1 else []

            # Get full qualified name
            full_name = (
                f"{self.current_namespace}.{class_name}"
                if self.current_namespace
                else class_name
            )

            # Get raw text
            raw_text = self._get_node_text_optimized(node)

            # Determine class type
            class_type_map = {
                "class_declaration": "class",
                "interface_declaration": "interface",
                "record_declaration": "record",
                "enum_declaration": "enum",
                "struct_declaration": "struct",
            }
            class_type = class_type_map.get(node.type, "class")

            return Class(
                name=class_name,
                start_line=node.start_point[0] + 1,
                end_line=node.end_point[0] + 1,
                raw_text=raw_text,
                full_qualified_name=full_name,
                superclass=superclass,
                interfaces=interfaces,
                modifiers=modifiers,
                visibility=visibility,
                annotations=attributes,
                class_type=class_type,
            )

        except Exception as e:
            log_error(f"Error extracting class declaration: {e}")
            return None

    def extract_functions(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Function]:
        """
        Extract methods, constructors, and properties.

        Args:
            tree: Tree-sitter AST tree parsed from C# source
            source_code: Original C# source code as string

        Returns:
            List of Function objects representing methods, constructors, and properties
        """
        self.source_code = source_code or ""
        self.content_lines = self.source_code.split("\n")
        self._reset_caches()

        functions: list[Function] = []

        if tree is None or tree.root_node is None:
            return functions

        # Extract namespace first
        self._extract_namespace(tree.root_node)

        # Extract methods, constructors, and properties
        for node in self._traverse_iterative(tree.root_node):
            if node.type == "method_declaration":
                func = self._extract_method(node)
                if func:
                    functions.append(func)
            elif node.type == "constructor_declaration":
                func = self._extract_constructor(node)
                if func:
                    functions.append(func)
            elif node.type == "property_declaration":
                func = self._extract_property(node)
                if func:
                    functions.append(func)

        # Sort by start line for deterministic output
        functions.sort(key=lambda f: f.start_line)

        return functions

    def _extract_method(self, node: "tree_sitter.Node") -> Function | None:
        """
        Extract a method declaration.

        Args:
            node: Method declaration node

        Returns:
            Function object or None if extraction fails
        """
        try:
            # Get method name
            name_node = node.child_by_field_name("name")
            if not name_node:
                return None

            method_name = self._get_node_text_optimized(name_node)

            # Get modifiers and visibility
            modifiers = self._extract_modifiers(node)
            visibility = self._determine_visibility(modifiers)

            # Check if async
            is_async = "async" in modifiers

            # Get attributes
            attributes = self._extract_attributes(node)

            # Get return type
            type_node = node.child_by_field_name("type")
            return_type = self._extract_type_name(type_node)

            # Get parameters
            params_node = node.child_by_field_name("parameters")
            parameters = self._extract_parameters(params_node)

            # Get raw text
            raw_text = self._get_node_text_optimized(node)

            # Calculate complexity (simplified)
            complexity_score = self._calculate_complexity(node)

            return Function(
                name=method_name,
                start_line=node.start_point[0] + 1,
                end_line=node.end_point[0] + 1,
                raw_text=raw_text,
                parameters=parameters,
                return_type=return_type,
                modifiers=modifiers,
                visibility=visibility,
                is_async=is_async,
                annotations=attributes,
                complexity_score=complexity_score,
            )

        except Exception as e:
            log_error(f"Error extracting method: {e}")
            return None

    def _extract_constructor(self, node: "tree_sitter.Node") -> Function | None:
        """
        Extract a constructor declaration.

        Args:
            node: Constructor declaration node

        Returns:
            Function object or None if extraction fails
        """
        try:
            # Get constructor name
            name_node = node.child_by_field_name("name")
            if not name_node:
                return None

            constructor_name = self._get_node_text_optimized(name_node)

            # Get modifiers and visibility
            modifiers = self._extract_modifiers(node)
            visibility = self._determine_visibility(modifiers)

            # Get attributes
            attributes = self._extract_attributes(node)

            # Get parameters
            params_node = node.child_by_field_name("parameters")
            parameters = self._extract_parameters(params_node)

            # Get raw text
            raw_text = self._get_node_text_optimized(node)

            return Function(
                name=constructor_name,
                start_line=node.start_point[0] + 1,
                end_line=node.end_point[0] + 1,
                raw_text=raw_text,
                parameters=parameters,
                return_type="void",
                modifiers=modifiers,
                visibility=visibility,
                is_constructor=True,
                annotations=attributes,
            )

        except Exception as e:
            log_error(f"Error extracting constructor: {e}")
            return None

    def _extract_property(self, node: "tree_sitter.Node") -> Function | None:
        """
        Extract a property declaration.

        Args:
            node: Property declaration node

        Returns:
            Function object with is_property=True or None if extraction fails
        """
        try:
            # Get property name
            name_node = node.child_by_field_name("name")
            if not name_node:
                return None

            property_name = self._get_node_text_optimized(name_node)

            # Get modifiers and visibility
            modifiers = self._extract_modifiers(node)
            visibility = self._determine_visibility(modifiers)

            # Get attributes
            attributes = self._extract_attributes(node)

            # Get property type
            type_node = node.child_by_field_name("type")
            property_type = self._extract_type_name(type_node)

            # Get raw text
            raw_text = self._get_node_text_optimized(node)

            # Check for getter/setter (for future use)
            # has_getter = False
            # has_setter = False
            # for child in node.children:
            #     if child.type == "accessor_list":
            #         for accessor in child.children:
            #             if accessor.type == "get_accessor_declaration":
            #                 has_getter = True
            #             elif accessor.type == "set_accessor_declaration":
            #                 has_setter = True
            #             elif accessor.type == "init_accessor_declaration":
            #                 has_setter = True  # init is a special setter

            return Function(
                name=property_name,
                start_line=node.start_point[0] + 1,
                end_line=node.end_point[0] + 1,
                raw_text=raw_text,
                parameters=[],
                return_type=property_type,
                modifiers=modifiers,
                visibility=visibility,
                is_property=True,
                annotations=attributes,
            )

        except Exception as e:
            log_error(f"Error extracting property: {e}")
            return None

    def _calculate_complexity(self, node: "tree_sitter.Node") -> int:
        """
        Calculate cyclomatic complexity of a method.

        Args:
            node: Method node

        Returns:
            Complexity score (1 + number of decision points)
        """
        complexity = 1
        decision_keywords = {
            "if_statement",
            "switch_statement",
            "for_statement",
            "foreach_statement",
            "while_statement",
            "do_statement",
            "catch_clause",
            "conditional_expression",  # ternary operator
        }

        for child in self._traverse_iterative(node):
            if child.type in decision_keywords:
                complexity += 1

        return complexity

    def extract_variables(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Variable]:
        """
        Extract fields, constants, and events.

        Args:
            tree: Tree-sitter AST tree parsed from C# source
            source_code: Original C# source code as string

        Returns:
            List of Variable objects representing fields
        """
        self.source_code = source_code or ""
        self.content_lines = self.source_code.split("\n")
        self._reset_caches()

        variables: list[Variable] = []

        if tree is None or tree.root_node is None:
            return variables

        # Extract fields
        for node in self._traverse_iterative(tree.root_node):
            if node.type == "field_declaration":
                vars_list = self._extract_field(node)
                variables.extend(vars_list)
            elif node.type == "event_field_declaration":
                vars_list = self._extract_event(node)
                variables.extend(vars_list)

        # Sort by start line for deterministic output
        variables.sort(key=lambda v: v.start_line)

        return variables

    def _extract_field(self, node: "tree_sitter.Node") -> list[Variable]:
        """
        Extract field declarations.

        Args:
            node: Field declaration node

        Returns:
            List of Variable objects (can be multiple if multiple variables declared)
        """
        variables: list[Variable] = []

        try:
            # Get modifiers
            modifiers = self._extract_modifiers(node)
            visibility = self._determine_visibility(modifiers)

            # Check if constant or readonly
            is_constant = "const" in modifiers
            # is_readonly = "readonly" in modifiers  # For future use

            # Get attributes
            attributes = self._extract_attributes(node)

            # Get field type
            type_node = None
            for child in node.children:
                if child.type == "variable_declaration":
                    type_node = child.child_by_field_name("type")
                    break

            field_type = self._extract_type_name(type_node)

            # Get variable declarators
            for child in node.children:
                if child.type == "variable_declaration":
                    for declarator in child.children:
                        if declarator.type == "variable_declarator":
                            name_node = declarator.child_by_field_name("name")
                            if name_node:
                                field_name = self._get_node_text_optimized(name_node)
                                raw_text = self._get_node_text_optimized(node)

                                variables.append(
                                    Variable(
                                        name=field_name,
                                        start_line=node.start_point[0] + 1,
                                        end_line=node.end_point[0] + 1,
                                        raw_text=raw_text,
                                        variable_type=field_type,
                                        modifiers=modifiers,
                                        visibility=visibility,
                                        is_constant=is_constant,
                                        annotations=attributes,
                                    )
                                )

        except Exception as e:
            log_error(f"Error extracting field: {e}")

        return variables

    def _extract_event(self, node: "tree_sitter.Node") -> list[Variable]:
        """
        Extract event field declarations.

        Args:
            node: Event field declaration node

        Returns:
            List of Variable objects representing events
        """
        variables: list[Variable] = []

        try:
            # Get modifiers
            modifiers = self._extract_modifiers(node)
            modifiers.append("event")  # Mark as event
            visibility = self._determine_visibility(modifiers)

            # Get attributes
            attributes = self._extract_attributes(node)

            # Get event type
            type_node = None
            for child in node.children:
                if child.type == "variable_declaration":
                    type_node = child.child_by_field_name("type")
                    break

            event_type = self._extract_type_name(type_node)

            # Get variable declarators
            for child in node.children:
                if child.type == "variable_declaration":
                    for declarator in child.children:
                        if declarator.type == "variable_declarator":
                            name_node = declarator.child_by_field_name("name")
                            if name_node:
                                event_name = self._get_node_text_optimized(name_node)
                                raw_text = self._get_node_text_optimized(node)

                                variables.append(
                                    Variable(
                                        name=event_name,
                                        start_line=node.start_point[0] + 1,
                                        end_line=node.end_point[0] + 1,
                                        raw_text=raw_text,
                                        variable_type=event_type,
                                        modifiers=modifiers,
                                        visibility=visibility,
                                        annotations=attributes,
                                    )
                                )

        except Exception as e:
            log_error(f"Error extracting event: {e}")

        return variables

    def extract_imports(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Import]:
        """
        Extract using directives.

        Args:
            tree: Tree-sitter AST tree parsed from C# source
            source_code: Original C# source code as string

        Returns:
            List of Import objects representing using directives
        """
        self.source_code = source_code or ""
        self.content_lines = self.source_code.split("\n")

        imports: list[Import] = []

        if tree is None or tree.root_node is None:
            return imports

        # Extract using directives
        for node in self._traverse_iterative(tree.root_node):
            if node.type == "using_directive":
                import_obj = self._extract_using_directive(node)
                if import_obj:
                    imports.append(import_obj)

        # Sort by start line for deterministic output
        imports.sort(key=lambda i: i.start_line)

        return imports

    def _extract_using_directive(self, node: "tree_sitter.Node") -> Import | None:
        """
        Extract a using directive.

        Args:
            node: Using directive node

        Returns:
            Import object or None if extraction fails
        """
        try:
            # Get the namespace or type being imported
            name_node = node.child_by_field_name("name")
            if not name_node:
                # Try to find qualified_name or identifier
                for child in node.children:
                    if child.type in ["qualified_name", "identifier", "name_equals"]:
                        name_node = child
                        break

            if not name_node:
                return None

            import_name = self._get_node_text_optimized(name_node)

            # Check if it's a static using
            is_static = False
            for child in node.children:
                if (
                    child.type == "static"
                    or self._get_node_text_optimized(child) == "static"
                ):
                    is_static = True
                    break

            # Get raw text
            raw_text = self._get_node_text_optimized(node)

            return Import(
                name=import_name,
                start_line=node.start_point[0] + 1,
                end_line=node.end_point[0] + 1,
                raw_text=raw_text,
                module_name=import_name,
                is_static=is_static,
            )

        except Exception as e:
            log_error(f"Error extracting using directive: {e}")
            return None


class CSharpPlugin(LanguagePlugin):
    """
    C# language plugin implementation.

    This plugin provides C# language support for tree-sitter-analyzer,
    enabling analysis of C# source files including modern C# features
    like records, nullable reference types, and async/await patterns.
    """

    def __init__(self) -> None:
        """Initialize the C# plugin."""
        super().__init__()
        self.extractor = CSharpElementExtractor()
        self.language = "csharp"
        self.supported_extensions = [".cs"]
        self._cached_language: Any | None = None

    def get_language_name(self) -> str:
        """
        Get the language name.

        Returns:
            Language name as string: "csharp"
        """
        return "csharp"

    def get_file_extensions(self) -> list[str]:
        """
        Get supported file extensions.

        Returns:
            List of file extensions: [".cs"]
        """
        return [".cs"]

    def create_extractor(self) -> ElementExtractor:
        """
        Create a new C# element extractor instance.

        Returns:
            CSharpElementExtractor instance
        """
        return CSharpElementExtractor()

    def get_queries(self) -> dict[str, str]:
        """
        Return C#-specific tree-sitter queries.

        Returns:
            Dictionary of query names to query strings
        """
        from ..queries.csharp import CSHARP_QUERIES

        return CSHARP_QUERIES

    def execute_query_strategy(
        self, query_key: str | None, language: str
    ) -> str | None:
        """
        Execute query strategy for C#.

        Args:
            query_key: Query key to execute
            language: Language name

        Returns:
            Query string or None if not applicable
        """
        if language != "csharp":
            return None

        queries = self.get_queries()
        return queries.get(query_key) if query_key else None

    def get_element_categories(self) -> dict[str, list[str]]:
        """
        Return C# element categories for query execution.

        Returns:
            Dictionary of category names to element types
        """
        return {
            "classes": ["class", "interface", "record", "enum", "struct"],
            "methods": ["method", "constructor"],
            "properties": ["property", "auto_property", "computed_property"],
            "fields": ["field", "const_field", "readonly_field", "event"],
            "imports": ["using", "static_using"],
            "attributes": ["attribute", "http_attribute", "authorize_attribute"],
            "async": ["async_method"],
            "linq": ["linq_query", "from_clause", "where_clause", "select_clause"],
            "control_flow": [
                "if_statement",
                "for_statement",
                "foreach_statement",
                "while_statement",
                "switch_statement",
                "try_statement",
            ],
        }

    def get_tree_sitter_language(self) -> Any | None:
        """
        Load tree-sitter-c-sharp language.

        Returns:
            Tree-sitter Language object or None if loading fails
        """
        if self._cached_language is not None:
            return self._cached_language

        try:
            import tree_sitter_c_sharp

            lang = tree_sitter_c_sharp.language()

            # Handle both old and new tree-sitter API
            if hasattr(lang, "__class__") and "Language" in str(type(lang)):
                self._cached_language = lang
            else:
                self._cached_language = tree_sitter.Language(lang)

            log_debug("Successfully loaded tree-sitter-c-sharp language")
            return self._cached_language

        except ImportError as e:
            log_error(f"tree-sitter-c-sharp not available: {e}")
            log_error("Install with: pip install tree-sitter-c-sharp")
            return None
        except Exception as e:
            log_error(f"Failed to load tree-sitter language for C#: {e}")
            return None

    async def analyze_file(
        self, file_path: str, request: "AnalysisRequest"
    ) -> "AnalysisResult":
        """
        Analyze a C# file and extract all elements.

        Args:
            file_path: Path to the C# file to analyze
            request: Analysis request containing file options

        Returns:
            AnalysisResult with extracted elements
        """
        from ..encoding_utils import read_file_safe
        from ..models import AnalysisResult

        try:
            # Read file content
            source_code, encoding = read_file_safe(file_path)
            self.extractor._file_encoding = encoding

            # Get tree-sitter language
            language = self.get_tree_sitter_language()
            if not language:
                log_error("Failed to load C# language")
                return AnalysisResult(
                    file_path=file_path,
                    language="csharp",
                    elements=[],
                    success=False,
                    error_message="Failed to load C# language",
                )

            # Parse the source code
            parser = tree_sitter.Parser()

            # Set language using the appropriate method
            if hasattr(parser, "set_language"):
                parser.set_language(language)
            elif hasattr(parser, "language"):
                parser.language = language
            else:
                parser = tree_sitter.Parser(language)

            tree = parser.parse(source_code.encode("utf-8"))

            # Extract all elements
            classes = self.extractor.extract_classes(tree, source_code)
            functions = self.extractor.extract_functions(tree, source_code)
            variables = self.extractor.extract_variables(tree, source_code)
            imports = self.extractor.extract_imports(tree, source_code)

            # Combine all elements into a single list
            elements: list[Any] = []
            elements.extend(classes)
            elements.extend(functions)
            elements.extend(variables)
            elements.extend(imports)

            # Count AST nodes
            node_count = sum(
                1 for _ in self.extractor._traverse_iterative(tree.root_node)
            )

            # Count lines
            line_count = len(source_code.split("\n"))

            return AnalysisResult(
                file_path=file_path,
                language="csharp",
                elements=elements,
                node_count=node_count,
                line_count=line_count,
                source_code=source_code,
            )

        except Exception as e:
            log_error(f"Error analyzing C# file {file_path}: {e}")
            return AnalysisResult(
                file_path=file_path,
                language="csharp",
                elements=[],
                success=False,
                error_message=str(e),
            )



#!/usr/bin/env python3
"""
Unified Analysis Engine - Common Analysis System for CLI and MCP (Fixed Version)

This module provides a unified engine that serves as the center of all analysis processing.
It is commonly used by CLI, MCP, and other interfaces.

Roo Code compliance:
- Type hints: Required for all functions
- MCP logging: Log output at each step
- docstring: Google Style docstring
- Performance-focused: Singleton pattern and cache sharing
"""

import hashlib
import threading
from dataclasses import dataclass
from typing import Any, Optional, Protocol

from ..models import AnalysisResult
from ..plugins.base import LanguagePlugin as BaseLanguagePlugin
from ..plugins.manager import PluginManager
from ..security import SecurityValidator
from ..utils import log_debug, log_error, log_info, log_performance
from .cache_service import CacheService


class UnsupportedLanguageError(Exception):
    """Unsupported language error"""

    pass


class PluginRegistry(Protocol):
    """Protocol for plugin registration management"""

    def get_plugin(self, language: str) -> Optional["LanguagePlugin"]:
        """Get language plugin"""
        ...


class LanguagePlugin(Protocol):
    """Language plugin protocol"""

    async def analyze_file(
        self, file_path: str, request: "AnalysisRequest"
    ) -> AnalysisResult:
        """File analysis"""
        ...


class PerformanceMonitor:
    """Performance monitoring (simplified version)"""

    def __init__(self) -> None:
        self._last_duration: float = 0.0
        self._monitoring_active: bool = False
        self._operation_stats: dict[str, Any] = {}
        self._total_operations: int = 0

    def measure_operation(self, operation_name: str) -> "PerformanceContext":
        """Return measurement context for operation"""
        return PerformanceContext(operation_name, self)

    def get_last_duration(self) -> float:
        """Get last operation time"""
        return self._last_duration

    def _set_duration(self, duration: float) -> None:
        """Set operation time (internal use)"""
        self._last_duration = duration

    def start_monitoring(self) -> None:
        """Start performance monitoring"""
        self._monitoring_active = True
        log_info("Performance monitoring started")

    def stop_monitoring(self) -> None:
        """Stop performance monitoring"""
        self._monitoring_active = False
        log_info("Performance monitoring stopped")

    def get_operation_stats(self) -> dict[str, Any]:
        """Get operation statistics"""
        return self._operation_stats.copy()

    def get_performance_summary(self) -> dict[str, Any]:
        """Get performance summary"""
        return {
            "total_operations": self._total_operations,
            "monitoring_active": self._monitoring_active,
            "last_duration": self._last_duration,
            "operation_count": len(self._operation_stats),
        }

    def record_operation(self, operation_name: str, duration: float) -> None:
        """Record operation"""
        if self._monitoring_active:
            if operation_name not in self._operation_stats:
                self._operation_stats[operation_name] = {
                    "count": 0,
                    "total_time": 0.0,
                    "avg_time": 0.0,
                    "min_time": float("inf"),
                    "max_time": 0.0,
                }

            stats = self._operation_stats[operation_name]
            stats["count"] += 1
            stats["total_time"] += duration
            stats["avg_time"] = stats["total_time"] / stats["count"]
            stats["min_time"] = min(stats["min_time"], duration)
            stats["max_time"] = max(stats["max_time"], duration)

            self._total_operations += 1

    def clear_metrics(self) -> None:
        """Clear collected metrics"""
        self._operation_stats.clear()
        self._total_operations = 0
        self._last_duration = 0.0
        log_info("Performance metrics cleared")


class PerformanceContext:
    """Performance measurement context"""

    def __init__(self, operation_name: str, monitor: PerformanceMonitor) -> None:
        self.operation_name = operation_name
        self.monitor = monitor
        self.start_time: float = 0.0

    def __enter__(self) -> "PerformanceContext":
        import time

        self.start_time = time.time()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        import time

        duration = time.time() - self.start_time
        self.monitor._set_duration(duration)
        self.monitor.record_operation(self.operation_name, duration)
        log_performance(self.operation_name, duration, "Operation completed")


@dataclass(frozen=True)
class AnalysisRequest:
    """
    Analysis request

    Attributes:
        file_path: Path to target file to analyze
        language: Programming language (auto-detected if None)
        include_complexity: Whether to include complexity metrics
        include_details: Whether to include detailed structure info
        format_type: Output format
    """

    file_path: str
    language: str | None = None
    include_complexity: bool = True
    include_details: bool = False
    format_type: str = "json"

    @classmethod
    def from_mcp_arguments(cls, arguments: dict[str, Any]) -> "AnalysisRequest":
        """
        Create analysis request from MCP tool arguments

        Args:
            arguments: MCP argument dictionary

        Returns:
            AnalysisRequest
        """
        return cls(
            file_path=arguments.get("file_path", ""),
            language=arguments.get("language"),
            include_complexity=arguments.get("include_complexity", True),
            include_details=arguments.get("include_details", False),
            format_type=arguments.get("format_type", "json"),
        )


# SimplePluginRegistry removed - now using PluginManager


class UnifiedAnalysisEngine:
    """
    Unified analysis engine (revised)

    Central engine shared by CLI, MCP and other interfaces, implemented as a
    singleton to enable efficient resource usage and cache sharing.

    Improvements:
    - Fix async issues in destructor
    - Provide explicit cleanup() method

    Attributes:
        _cache_service: Cache service
        _plugin_manager: Plugin manager
        _performance_monitor: Performance monitor
    """

    _instances: dict[str, "UnifiedAnalysisEngine"] = {}
    _lock: threading.Lock = threading.Lock()

    def __new__(cls, project_root: str | None = None) -> "UnifiedAnalysisEngine":
        """Singleton instance sharing (project_root aware)"""
        # Create a key based on project_root for different instances
        instance_key = project_root or "default"

        if instance_key not in cls._instances:
            with cls._lock:
                if instance_key not in cls._instances:
                    instance = super().__new__(cls)
                    cls._instances[instance_key] = instance
                    # Mark as not initialized for this instance
                    instance._initialized = False

        return cls._instances[instance_key]

    def __init__(self, project_root: str | None = None) -> None:
        """Initialize (executed only once per instance)"""
        if hasattr(self, "_initialized") and getattr(self, "_initialized", False):
            return

        self._cache_service = CacheService()
        self._plugin_manager = PluginManager()
        self._performance_monitor = PerformanceMonitor()
        self._security_validator = SecurityValidator(project_root)
        self._project_root = project_root

        # Auto-load plugins
        self._load_plugins()

        self._initialized = True

        log_debug(
            f"UnifiedAnalysisEngine initialized with project root: {project_root}"
        )

    def _load_plugins(self) -> None:
        """Auto-load available plugins"""
        log_debug("Loading plugins using PluginManager...")

        try:
            # PluginManagerの自動ロード機能を使用
            loaded_plugins = self._plugin_manager.load_plugins()

            final_languages = [plugin.get_language_name() for plugin in loaded_plugins]
            log_debug(
                f"Successfully loaded {len(final_languages)} language plugins: {', '.join(final_languages)}"
            )
        except Exception as e:
            log_error(f"Failed to load plugins: {e}")
            import traceback

            log_error(f"Plugin loading traceback: {traceback.format_exc()}")

    async def analyze(self, request: AnalysisRequest) -> AnalysisResult:
        """
        Unified analysis method

        Args:
            request: Analysis request

        Returns:
            Analysis result

        Raises:
            UnsupportedLanguageError: When language is not supported
            FileNotFoundError: When file is not found
        """
        log_debug(f"Starting analysis for {request.file_path}")

        # Security validation
        is_valid, error_msg = self._security_validator.validate_file_path(
            request.file_path
        )
        if not is_valid:
            log_error(
                f"Security validation failed for file path: {request.file_path} - {error_msg}"
            )
            raise ValueError(f"Invalid file path: {error_msg}")

        # Cache check (shared across CLI/MCP)
        cache_key = self._generate_cache_key(request)
        cached_result = await self._cache_service.get(cache_key)
        if cached_result:
            log_info(f"Cache hit for {request.file_path}")
            return cached_result  # type: ignore

        # Language detection
        language = request.language or self._detect_language(request.file_path)
        log_debug(f"Detected language: {language}")

        # Debug: inspect registered plugins
        supported_languages = self._plugin_manager.get_supported_languages()
        log_debug(f"Supported languages: {supported_languages}")
        log_debug(f"Looking for plugin for language: {language}")

        # Get plugin
        plugin = self._plugin_manager.get_plugin(language)
        if not plugin:
            error_msg = f"Language {language} not supported"
            log_error(error_msg)
            raise UnsupportedLanguageError(error_msg)

        log_debug(f"Found plugin for {language}: {type(plugin)}")

        # Run analysis (with performance monitoring)
        with self._performance_monitor.measure_operation(f"analyze_{language}"):
            log_debug(f"Calling plugin.analyze_file for {request.file_path}")
            result = await plugin.analyze_file(request.file_path, request)
            log_debug(
                f"Plugin returned result: success={result.success}, elements={len(result.elements) if result.elements else 0}"
            )

        # Ensure language field is set
        if result.language == "unknown" or not result.language:
            result.language = language

        # Save to cache
        await self._cache_service.set(cache_key, result)

        log_performance(
            "unified_analysis",
            self._performance_monitor.get_last_duration(),
            f"Analyzed {request.file_path} ({language})",
        )

        return result

    async def analyze_file(self, file_path: str) -> AnalysisResult:
        """
        Backward compatibility method for analyze_file.

        Args:
            file_path: Path to the file to analyze

        Returns:
            Analysis result
        """
        # Security validation
        is_valid, error_msg = self._security_validator.validate_file_path(file_path)
        if not is_valid:
            log_error(
                f"Security validation failed for file path: {file_path} - {error_msg}"
            )
            raise ValueError(f"Invalid file path: {error_msg}")

        request = AnalysisRequest(
            file_path=file_path,
            language=None,  # Auto-detect
            include_complexity=True,
            include_details=True,
        )
        return await self.analyze(request)

    def _generate_cache_key(self, request: AnalysisRequest) -> str:
        """
        Generate cache key

        Args:
            request: Analysis request

        Returns:
            Hashed cache key
        """
        # 一意なキーを生成するための文字列を構築
        key_components = [
            request.file_path,
            str(request.language),
            str(request.include_complexity),
            str(request.include_details),
            request.format_type,
        ]

        key_string = ":".join(key_components)

        # SHA256でハッシュ化
        return hashlib.sha256(key_string.encode("utf-8")).hexdigest()

    def _detect_language(self, file_path: str) -> str:
        """
        Detect language from file extension

        Args:
            file_path: File path

        Returns:
            Detected language name
        """
        # 簡易的な拡張子ベース検出
        from pathlib import Path

        ext = Path(file_path).suffix

        language_map = {
            ".java": "java",
            ".py": "python",
            ".js": "javascript",
            ".ts": "typescript",
            ".c": "c",
            ".cpp": "cpp",
            ".cc": "cpp",
            ".cxx": "cpp",
            ".rs": "rust",
            ".go": "go",
            ".sql": "sql",
        }

        detected = language_map.get(ext.lower(), "unknown")
        log_debug(f"Language detection: {file_path} -> {detected}")
        return detected

    def clear_cache(self) -> None:
        """Clear cache (for tests)"""
        self._cache_service.clear()
        log_info("Analysis engine cache cleared")

    def register_plugin(self, language: str, plugin: BaseLanguagePlugin) -> None:
        """
        Register plugin

        Args:
            language: Language name (kept for compatibility, not used)
            plugin: Language plugin instance
        """
        self._plugin_manager.register_plugin(plugin)

    def get_supported_languages(self) -> list[str]:
        """
        Get list of supported languages

        Returns:
            List of language names
        """
        return self._plugin_manager.get_supported_languages()

    def get_cache_stats(self) -> dict[str, Any]:
        """
        Get cache statistics

        Returns:
            Cache statistics dictionary
        """
        return self._cache_service.get_stats()

    async def invalidate_cache_pattern(self, pattern: str) -> int:
        """
        Invalidate cached entries matching a pattern

        Args:
            pattern: Pattern to match keys

        Returns:
            Number of invalidated keys
        """
        return await self._cache_service.invalidate_pattern(pattern)

    def measure_operation(self, operation_name: str) -> "PerformanceContext":
        """
        Context manager for performance measurement

        Args:
            operation_name: Operation name

        Returns:
            PerformanceContext
        """
        return self._performance_monitor.measure_operation(operation_name)

    def start_monitoring(self) -> None:
        """Start performance monitoring"""
        self._performance_monitor.start_monitoring()

    def stop_monitoring(self) -> None:
        """Stop performance monitoring"""
        self._performance_monitor.stop_monitoring()

    def get_operation_stats(self) -> dict[str, Any]:
        """Get operation statistics"""
        return self._performance_monitor.get_operation_stats()

    def get_performance_summary(self) -> dict[str, Any]:
        """Get performance summary"""
        return self._performance_monitor.get_performance_summary()

    def clear_metrics(self) -> None:
        """
        Clear collected performance metrics

        Resets metrics collected by performance monitoring. Used in tests/debugging.
        """
        # 新しいパフォーマンスモニターインスタンスを作成してリセット
        self._performance_monitor = PerformanceMonitor()
        log_info("Performance metrics cleared")

    def cleanup(self) -> None:
        """
        Explicit resource cleanup

        Call explicitly (e.g., at end of tests) to clean up resources and avoid
        async issues in destructors.
        """
        try:
            if hasattr(self, "_cache_service"):
                self._cache_service.clear()
            if hasattr(self, "_performance_monitor"):
                self._performance_monitor.clear_metrics()
            log_debug("UnifiedAnalysisEngine cleaned up")
        except Exception as e:
            log_error(f"Error during UnifiedAnalysisEngine cleanup: {e}")

    def __del__(self) -> None:
        """
        Destructor - keep minimal to avoid issues in async contexts

        Performs no cleanup; use cleanup() explicitly when needed.
        """
        # デストラクタでは何もしない（非同期コンテキストでの問題を避けるため）
        pass


# 簡易的なプラグイン実装（テスト用）
class MockLanguagePlugin:
    """Mock plugin for testing"""

    def __init__(self, language: str) -> None:
        self.language = language

    def get_language_name(self) -> str:
        """Get language name"""
        return self.language

    def get_file_extensions(self) -> list[str]:
        """Get supported file extensions"""
        return [f".{self.language}"]

    def create_extractor(self) -> None:
        """Create extractor (mock)"""
        return None

    async def analyze_file(
        self, file_path: str, request: AnalysisRequest
    ) -> AnalysisResult:
        """Mock analysis implementation"""
        log_info(f"Mock analysis for {file_path} ({self.language})")

        # 簡易的な解析結果を返す
        return AnalysisResult(
            file_path=file_path,
            line_count=10,  # 新しいアーキテクチャ用
            elements=[],  # 新しいアーキテクチャ用
            node_count=5,  # 新しいアーキテクチャ用
            query_results={},  # 新しいアーキテクチャ用
            source_code="// Mock source code",  # 新しいアーキテクチャ用
            language=self.language,  # 言語を設定
            package=None,
            analysis_time=0.1,
            success=True,
            error_message=None,
        )


def get_analysis_engine(project_root: str | None = None) -> UnifiedAnalysisEngine:
    """
    Get unified analysis engine instance

    Args:
        project_root: Project root directory for security validation

    Returns:
        Singleton instance of UnifiedAnalysisEngine
    """
    return UnifiedAnalysisEngine(project_root)
